"""
Functions for reading and writing Agilent/Varian binary (fid) files and
parameter (procpar) files.
"""

from __future__ import print_function, division

__developer_doc__ = """
Agilent/Varian file format information
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
Both the Agilent/Varian binary and parameter formats are documented in:

* Varian MR News 2005-04-18 Importing Binary VnmrJ / VNMR FIDs into Third
    Party Software and VnmrJ / VNMR FID Data Format
* VnmrJ User Programming - Chapter 5: Parameters and Data

These are available (as of 04/2011) online from
`Agilent <http://agilent.com>`_.

"""

import os
import struct
import inspect
from warnings import warn
from functools import reduce

import numpy as np

from . import fileiobase

############################
# dictionary/data creation #
############################


def create_data(data):
    """
    Create a Agilent/Varian data array (recast into complex64 array)
    """
    return np.array(data, dtype="complex64")

########################
# universal dictionary #
########################


def guess_udic(dic, data):
    """
    Guess parameter of a universal dictionary from dic, data pair.

    Parameters
    ----------
    dic : dict
        Dictionary of Agilent/Varian parameters.
    data : ndarray
        Array of NMR data.

    Returns
    -------
    udic : dict
        Universal dictionary of spectral parameters.

    """
    # create an empty universal dictionary
    udic = fileiobase.create_blank_udic(data.ndim)

    # update default values
    for i in range(data.ndim):
        udic[i]["size"] = data.shape[i]

    return udic


def create_dic(udic):
    """
    Create a Agilent/Varian parameter dictionary from a universal dictionary.

    Parameters
    ----------
    udic : dict
        Universal dictionary of spectral parameters.

    Returns
    -------
    dic : dict
        Dictionary of Agilent/Varian parameters
    """
    ddic = udic[udic["ndim"] - 1]   # direct dimension dictionary

    # number of points in direct dimension is R+I
    if ddic["complex"]:
        num_points = ddic["size"] * 2
    else:
        num_points = ddic["size"]

    # number of blocks is product of R+I of all other dimensions
    num_blocks = 1
    for f in [udic[k]["size"] for k in range(udic["ndim"] - 1)]:
        num_blocks = num_blocks * f

    # create the varian dictionary
    dic = dict()

    # default flag bits
    dic["S_DATA"] = 1
    dic["S_SPEC"] = 0
    dic["S_32"] = 0
    dic["S_FLOAT"] = 1
    dic["S_COMPLEX"] = 0
    dic["S_HYPERCOMPLEX"] = 0
    dic["S_ACQPAR"] = 1
    dic["S_SECND"] = 0
    dic["S_TRANSF"] = 0
    dic["S_NP"] = 0
    dic["S_NF"] = 0
    dic["S_NI"] = 0
    dic["S_NI2"] = 0

    # parameters
    dic["nblocks"] = num_blocks
    dic["ntraces"] = 1
    dic["np"] = num_points
    dic["ebytes"] = 4
    dic["tbytes"] = 4 * num_points
    dic["bbytes"] = dic["tbytes"] + 28
    dic["vers_id"] = 0
    dic["status"] = 201     # 6th bit set
    dic["nbheaders"] = 1

    # fake procpar dictionary
    pdic = dict()

    # direct dimension
    ndim = udic["ndim"]
    pdic["np"] = create_pdic_param('np', [str(udic[ndim - 1]['size'] * 2)])

    if ndim >= 2:
        if udic[ndim - 2]['complex']:
            pdic["ni"] = create_pdic_param('ni',
                                           [str(udic[ndim - 2]['size'] // 2)])
            pdic["phase"] = create_pdic_param('phase', ['0', '1'])
        else:
            pdic["ni"] = create_pdic_param('ni', [str(udic[ndim - 2]['size'])])
            pdic["phase"] = create_pdic_param('phase', ['0'])

    if ndim >= 3:
        pdic['array'] = create_pdic_param('array', ['phase,phase2'])
        if udic[ndim - 3]['complex']:
            pdic["ni2"] = create_pdic_param(
                'ni2', [str(udic[ndim - 3]['size'] // 2)])
            pdic["phase2"] = create_pdic_param('phase', ['0', '1'])
        else:
            pdic["ni2"] = create_pdic_param('ni2',
                                            [str(udic[ndim - 3]['size'])])
            pdic["phase2"] = create_pdic_param('phase2', ['0'])

    if ndim >= 4:
        pdic['array'] = create_pdic_param('array', ['phase,phase2,phase3'])
        if udic[ndim - 4]['complex']:
            pdic["ni3"] = create_pdic_param(
                'ni3', [str(udic[ndim - 4]['size'] // 2)])
            pdic["phase3"] = create_pdic_param('phase3', ['0', '1'])
        else:
            pdic["ni3"] = create_pdic_param(
                'ni3', [str(udic[ndim - 4]['size'])])
            pdic["phase3"] = create_pdic_param('phase3', ['0'])

    dic["procpar"] = pdic
    return dic


def create_pdic_param(name, values):
    """
    Create a fake procpar dictionary element of with given name and values.
    """
    dic = dict()
    dic["Dgroup"] = '1'
    dic["Ggroup"] = '2'
    dic["active"] = '1'
    dic["basictype"] = '1'
    dic["enumerable"] = '0'
    dic["intptr"] = '64'
    dic["maxvalue"] = '32767'   # 2^15-1
    dic["minvalue"] = '0'
    dic["name"] = name
    dic["protection"] = '0'
    dic["stepsize"] = '0'
    dic["subtype"] = '7'
    dic["values"] = values
    return dic

########################
# file reading/writing #
########################


def read(dir=".", fid_file="fid", procpar_file="procpar", read_blockhead=False,
         shape=None, torder=None, as_2d=False):
    """
    Read Agilent/Varian files in a directory.

    Parameters
    ----------
    dir : str, optional
        Directory holding Agilent/Varian data. Default is the current working
        directory.
    fid_file : str, optional
        Filename of binary (fid) file in directory.
    procpar_file : str, optional
        Filename of procpar file in directory.
    read_blockhead : bool, optional
        True to read blockheader(s) and return then in the Agilent/Varian
        parameter dictionary. False (default) does not perform this reading.
    shape : tuple of ints, optional
        Shape of data in binary file. None (default)  will attempt to finds
        this automatically.
    torder : {None, 'r', 'o', 'f' or a Python function} , optional
        Description of the mapping of traces in the file to the NMR data
        matrix. None (the default) will attempt to find this automatically
        which is typically fine for most NMR experiments. See below for
        additional details.
    as_2d : bool, optional
        True to return data as a 2D array ignorning the shape and torder
        parameters.

    Returns
    -------
    dic : dict
        Dictionary of Agilent/Varian parameters.
    data : ndarray
        Array of NMR data

    Notes
    -----
    The torder parameter describes how the traces on disk should be
    re-organized to form the NMR data matrix.  In most cases this can be
    determined automatically by examining the order of phase parameters in the
    procpar file.  This is done if torder is set to None.  In some cases
    it must be explicitly provided.  Three common cases are:

    ======== ======================  ================= =======================
    Name     Ordering of arrays      torder            Notes
    ======== ======================  ================= =======================
    regular  d3, d2, phase2, phase   'regular' or 'r'
    opposite d3, d2, phase, phase2   'opposite' or 'o'
    flat     As data exists in file  'flat' or 'f'     Valid for 1D or 2D data
    ======== ======================  ================= =======================

    In addition a function which maps indirect dimension index tuples to/from
    trace numbers as stored on disk can be provided.  For reading this
    function should take 2 arguments: shape, index_tuple and return an integer
    trace number.  For writing this function should again take 2 arguments:
    shape,trace_number and return the indirect dimension index tuple for the
    given trace.

    See Also
    --------
    read_lowmem : Read Agilent/Varian files using mimimal amounts of memory.
    write : Write Agilent/Varian files.

    """
    if os.path.isdir(dir) is False:
        raise IOError("directory %s does not exist" % (dir))

    # read in the procpar file
    pdic = read_procpar(os.path.join(dir, procpar_file))

    # determine shape and torder is needed
    if as_2d is False:
        if shape is None:
            shape = find_shape(pdic)
        if torder is None and shape is not None and len(shape) >= 3:
            torder = find_torder(pdic, shape)

    # read in the fid file
    fname = os.path.join(dir, fid_file)
    dic, data = read_fid(fname, shape, torder, as_2d, read_blockhead)

    # add the procpar dictionary to the main dictionary
    dic["procpar"] = pdic

    return dic, data


def read_lowmem(dir=".", fid_file="fid", procpar_file="procpar",
                read_blockhead=False, shape=None, torder=None):
    """
    Read Agilent/Varian files in a directory using minimal amount of memory.

    Parameters
    ----------
    dir : str, optional
        Directory holding Agilent/Varian data. Default is the current working
        directory.
    fid_file : str, optional
        Filename of binary (fid) file in directory.
    procpar_file : str, optional
        Filename of procpar file in directory.
    read_blockhead : bool, optional
        True to read blockheader(s) and return then in the Agilent/Varian
        parameter dictionary. False (default) does not perform this reading.
    shape : tuple of ints, optional
        Shape of data in binary file. None (default)  will attempt to finds
        this automatically.
    torder : {None, 'r', 'o', 'f' or a Python function} , optional
        Description of the mapping of traces in the file to the NMR data
        matrix. None (the default) will attempt to find this automatically
        which is typically fine for most NMR experiments. See :py:func:`read`
        for additional details.

    Returns
    -------
    dic : dict
        Dictionary of Agilent/Varian parameters.
    data : array_like
        Low memory object which can access NMR data on demand.

    See Also
    --------
    read : Read Agilent/Varian files.
    write_lowmem : Write Agilent/Varian files using mimimal memory

    """
    if os.path.isdir(dir) is False:
        raise IOError("directory %s does not exist" % (dir))

    # read in the procpar file
    pdic = read_procpar(os.path.join(dir, procpar_file))

    if shape is None:
        shape = find_shape(pdic)

    if torder is None:    # always try to find trace order
        torder = find_torder(pdic, shape)

    # read in the fid file
    fname = os.path.join(dir, fid_file)
    dic, data = read_fid_lowmem(fname, shape, torder, read_blockhead)

    # add the procpar dictionary to the main dictionary
    dic["procpar"] = pdic

    return dic, data


def write(dir, dic, data, fid_file="fid", procpar_file="procpar",
          torder=None, repack=False, overwrite=False):
    """
    Write Agilent/Varian files to a directory.

    Parameters
    ----------
    dir : str
        Name of the directory to write to.
    dic : dict
        Dictionary of Agilent/Varian parameters.
    data : array_like
        Array of NMR data to write.
    fid_file : str, optional
        Filename of binary (fid) file in directory to write to.
    procpar_file : str, optional
        Filename of procpar file in directory to write to.
    torder : {None, 'r', 'o', 'f' or a Python function} , optional
        Description of the mapping of traces in the file to the NMR data
        matrix. None (the default) will attempt to find this automatically
        which is typically fine for most NMR experiments. See :py:func:`read`
        for additional details.
    repack : bool, optional
        True to repack file and block headers. False (default) leaves these as
        is.
    overwrite : bool, optional
        Set to True to overwrite existing files, False will raise a Warning if
        files already exist.

    See Also
    --------
    write_lowmem : Write Agilent/Varian files using mimimal memory
    read : Read Agilent/Varian files.

    """
    if torder is None and data.ndim >= 3:
        torder = find_torder(dic["procpar"], data.shape)

    # write out the fid file
    fname = os.path.join(dir, fid_file)
    write_fid(fname, dic, data, torder, repack=repack, overwrite=overwrite)

    # write out procpar file
    write_procpar(os.path.join(dir, procpar_file), dic["procpar"], overwrite)

    return


def write_lowmem(dir, dic, data, fid_file="fid", procpar_file="procpar",
                 torder=None, repack=False, overwrite=False):
    """
    Write Agilent/Varian files to a directory using mimimal amounts of memory.

    Parameters
    ----------
    dir : str
        Name of the directory to write to.
    dic : dict
        Dictionary of Agilent/Varian parameters.
    data : array_like
        Array of NMR data to write.
    fid_file : str, optional
        Filename of binary (fid) file in directory to write to.
    procpar_file : str, optional
        Filename of procpar file in directory to write to.
    torder : {None, 'r', 'o', 'f' or a Python function} , optional
        Description of the mapping of traces in the file to the NMR data
        matrix. None (the default) will attempt to find this automatically
        which is typically fine for most NMR experiments. See :py:func:`read`
        for additional details.
    repack : bool, optional
        True to repack file and block headers. False (default) leaves these as
        is.
    overwrite : bool, optional
        Set to True to overwrite existing files, False will raise a Warning if
        files already exist.

    See Also
    --------
    write : Write Agilent/Varian files.
    read_lowmem : Read Agilent/Varian using minimal amounts of memory.

    """
    # always find trace ording
    if torder is None:
        torder = find_torder(dic["procpar"], data.shape)

    # write out the fid file
    fname = os.path.join(dir, fid_file)
    write_fid_lowmem(fname, dic, data, torder, repack=repack,
                     overwrite=overwrite)

    # write out procpar file
    write_procpar(os.path.join(dir, procpar_file), dic["procpar"], overwrite)

    return

############
# ordering #
############


def find_torder(dic, shape):
    """
    Find the torder from the procpar dictionary.

    If propar dictionary is incomplete a UserWarning is issued and 'r' is
    returned.

    Parameters
    ----------
    dic : dict
        Dictionary of parameters in the procpar file.
    shaoe : tuple of ints
        Shape of NMR data.

    Returns
    --------
    torder : {'r', 'f', 'o'}
        File ording for using in :py:func:`read` or :py:func:`write`.

    """
    ndim = len(shape)

    # 1 and 2D files are flat
    if ndim < 3:
        return 'f'  # flat

    if "array" not in dic:
        warn("no array in dictionary, torder set to 'r'")
        return 'r'

    # extract the array list
    al = dic['array']['values'][0].split(',')

    # remove one dimension if non-phase parameter is present in array list
    if False in [s.startswith('phase') for s in al]:
        ndim = ndim - 1
        if ndim < 3:
            return 'f'  # flat

    if ndim == 3:
        if "phase" in al and "phase2" in al:
            if al.index("phase") > al.index("phase2"):
                return 'r'  # regular
            else:
                return 'o'  # opposite
        else:
            warn("missing phase order, torder set to 'r'")
            return 'r'

    if ndim == 4:
        if "phase" in al and "phase2" in al and "phase3" in al:
            if al.index("phase") > al.index("phase2") > al.index("phase3"):
                return 'r'  # regular
            if al.index("phase") < al.index("phase2") < al.index("phase3"):
                return 'o'  # opposite
            else:
                warn("bad phase order, torder set to 'r'")
                return 'r'
        else:
            warn("missing phase order, torder set to 'r'")
            return 'r'

    warn("No trace ordering for" + str(ndim) +
         "dimensional data, torder set to 'r'")
    return 'r'


def torder2i2t(torder):
    """
    Convert torder to an index2trace function
    """
    # if torder is a function, return it
    if inspect.isfunction(torder):
        return torder
    elif torder == 'flat' or torder == 'f':
        return fileiobase.index2trace_flat
    elif torder == 'opposite' or torder == 'o':
        return fileiobase.index2trace_opp
    elif torder == 'regular' or torder == 'r':
        return fileiobase.index2trace_reg
    else:
        raise ValueError("unknown torder" + str(torder))


def torder2t2i(torder):
    """
    Convert torder to a trace2index functions
    """
    if inspect.isfunction(torder):
        return torder
    elif torder == 'flat' or torder == 'f':
        return fileiobase.trace2index_flat
    elif torder == 'opposite' or torder == 'o':
        return fileiobase.trace2index_opp
    elif torder == 'regular' or torder == 'r':
        return fileiobase.trace2index_reg
    else:
        raise ValueError("unknown torder" + str(torder))


def reorder_data(data, shape, torder):
    """
    Reorder raw data from file.

    Parameters
    ----------
    data : 2D ndarray
        Raw data as ordered in binary file.
    shape : tuple of ints
        Shape of the NMR data.
    torder : {'f', 'r', 'o' of Pytho function}
        Trace ordering . See :py:func:`read` for details.

    Returns
    -------
    tdata : ndarray
        Array of NMR data.

    Notes
    -----
    Minimal error checking is done to see if data and shape contain the same
    number of values.  This should be done before calling this function.

    """
    # take care of flat files...
    if torder == 'flat' or torder == 'f':
        try:
            data = data.reshape(shape)
        except ValueError:
            warn(str(data.shape) + "cannot be shaped into" + str(shape))
        return data

    # all other cases
    # make an empty array to hold reordered data
    ndata = np.empty(shape, dtype=data.dtype)

    # index2tuple converter
    i2t = torder2i2t(torder)

    # loop over all non-direct dimension indices
    for tup in np.ndindex(shape[:-1]):
        # determine the corresponding trace
        ntrace = i2t(shape[:-1], tup)
        # write the trace to the index
        ndata[tup] = data[ntrace]
    return ndata


def order_data(data, torder):
    """
    Order NMR data for writing to file.

    Parameters
    ----------
    data : ndarray
        Array of NMR data.
    torder : {'f', 'r', 'o'}
        Trace ordering.  See :py:func:`read` for details.

    Returns
    -------
    raw_data : 2D ndarray
        2D array ordered for writing to Agilent/Varian binary file.

    """
    # determine the shape of the on disk data matrix
    ntraces = reduce(lambda x, y: x * y, data.shape[:-1])
    nshape = (ntraces, data.shape[-1])

    # take care of flat files
    if torder == 'flat' or torder == 'f':
        return data.reshape(nshape)

    # make an emprt array to hold the 2D disk formated data matrix
    ndata = np.empty(nshape, dtype=data.dtype)

    # index2tuple converter
    t2i = torder2t2i(torder)

    # loop over all non-direct dimension indices
    for ntrace in range(ntraces):
        tup = t2i(data.shape[:-1], ntrace)
        ndata[ntrace] = data[tup]
    return ndata


#######################
# fid reading/writing #
#######################


def read_fid(filename, shape=None, torder='flat', as_2d=False,
             read_blockhead=False):
    """
    Read a Agilent/Varian binary (fid) file.

    Parameters
    ----------
    filename : str
        Filename of Agilent/Varian binary file (fid) to read.
    shape : tuple of ints, optional
        Shape of the binary data. If not provided data is returned as a 2D
        array. Required if more than one trace per block (non-standard).
    torder : {'f', 'n', 'o'}
        Trace order. See :py:func:`read` for details.
    as_2d : bool, optional
        True to return the data as a 2D array, ignoring the shape and torder
        parameters.
    read_blockhead : bool, optional
        True to read the Agilent/Varian blockheaders(s) into the returned
        dictionary. False ignores them.

    Returns
    -------
    dic : dict
        Dictionary of Agilent/Varian binary file parameters.
    data : ndarray
        Array of NMR data.

    See Also
    --------
    read_fid_lowmem : Read a Agilent/Varian binary file using minimal amounts
        of memory.
    read : Read Agilent/Varian files from a directory.

    """
    # open the file
    f = open(filename, 'rb')

    # read the fileheader
    dic = fileheader2dic(get_fileheader(f))

    # if ntraces is not 1 use _ntraces version
    if dic["ntraces"] != 1:
        return read_fid_ntraces(filename, shape, torder, as_2d, read_blockhead)

    # data parameters
    dt = find_dtype(dic)
    nblocks = dic["nblocks"]
    pts = dic["np"]
    nbheaders = dic["nbheaders"]

    # read the data
    if read_blockhead:
        bdic, data = get_nblocks(f, nblocks, pts, nbheaders, dt,
                                 read_blockhead)
        dic["blockheader"] = bdic
    else:
        data = get_nblocks(f, nblocks, pts, nbheaders, dt, read_blockhead)
    f.close()

    # uninterleave the real and imaginary data
    data = uninterleave_data(data)

    # return as 2D is requested
    if as_2d:
        return dic, data

    # return raw data if no shape provided.
    if shape is None:
        warn("unknown shape, returning unshaped data")
        return dic, data

    # check for 1D
    if data.shape[0] == 1:
        return dic, np.squeeze(data)

    # try to reshape
    # reorder 3D/4D data
    if len(shape) >= 3:
        try:
            return dic, reorder_data(data, shape, torder)
        except:
            warn("data cannot be re-ordered, returning raw 2D data\n" +
                 "Provided shape: " + str(shape) + " torder: " + str(torder))
            return dic, data

    try:
        data = data.reshape(shape)
    except ValueError:
        warn(str(data.shape) + "cannot be shaped into" + str(shape))
        return dic, data

    return dic, data


def read_fid_lowmem(filename, shape=None, torder='flat', as_2d=False,
                    read_blockhead=False):
    """
    Read a Agilent/Varian binary (fid) file using mimimal amounts of memory.

    Parameters
    ----------
    filename : str
        Filename of Agilent/Varian binary file (fid) to read.
    shape : tuple of ints, optional
        Shape of the binary data. If not provided data is returned as a 2D
        array. Required if more than one trace per block (non-standard).
    torder : {'f', 'n', 'o'}
        Trace order. See :py:func:`read` for details.
    read_blockhead : bool, optional
        Not used, retained for compatibility with :py:func:`read_fid`

    Returns
    -------
    dic : dict
        Dictionary of Agilent/Varian binary file parameters.
    data : array_like
        Low memory object which can access NMR data on demand.

    See Also
    --------
    read_fid : Read a Agilent/Varian binary file.
    read_lowmem : Read Agilent/Varian files from a directory using minimal
        amounts of memory.

    """
    # open the file
    f = open(filename, 'rb')

    # read the fileheader
    dic = fileheader2dic(get_fileheader(f))
    f.close()

    if dic["ntraces"] != 1:
        raise NotImplementedError

    i2tfunc = torder2i2t(torder)
    data = fid_nd(filename, i2tfunc, shape)
    return dic, data


def read_fid_ntraces(filename, shape=None, torder='flat', as_2d=False,
                     read_blockhead=False):
    """
    Read a Agilent/Varian binary (fid) file possibility having multiple
    traces per block.

    Parameters
    ----------
    filename : str
        Filename of Agilent/Varian binary file (fid) to read.
    shape : tuple of ints, optional
        Shape of the binary data. If not provided data is returned as a 2D
        array. Required if more than one trace per block (non-standard).
    torder : {'f', 'n', 'o'}
        Trace order. See :py:func:`read` for details.
    as_2d : bool, optional
        True to return the data as a 2D array, ignoring the shape and torder
        parameters.
    read_blockhead : bool, optional
        True to read the Agilent/Varian blockheaders(s) into the returned
        dictionary. False ignores them.

    Returns
    -------
    dic : dict
        Dictionary of Agilent/Varian binary file parameters.
    data : array_like
        Low memory object which can access NMR data on demand.

    See Also
    --------
    read_fid : Read a Agilent/Varian binary file with one trace per block.
    read_fid_lowmem : Read a Agilent/Varian binary file with one trace per
        block using minimal amounts of memory.

    """
    # open the file
    f = open(filename, 'rb')

    # read the fileheader
    dic = fileheader2dic(get_fileheader(f))

    # data parameters
    dt = find_dtype(dic)
    nblocks = dic["nblocks"]
    pts = dic["np"]
    nbheaders = dic["nbheaders"]
    ntraces = dic["ntraces"]

    # read the data
    if read_blockhead:
        bdic, data = get_nblocks_ntraces(f, nblocks, ntraces, pts,
                                         nbheaders, dt, read_blockhead)
        dic["blockheader"] = bdic
    else:
        data = get_nblocks_ntraces(f, nblocks, ntraces, pts, nbheaders, dt,
                                   read_blockhead)
    f.close()

    # uninterleave the real and imaginary data
    data = uninterleave_data(data)

    # if 2D array requested, return unshaped
    if as_2d:
        return dic, data

    # check for 1D
    if data.shape[0] == 1:
        return dic, np.squeeze(data)

    # try to reshape
    if shape is None:
        warn("unknown shape, returning unshaped data")
        return dic, data

    # reorder 3D/4D data
    if len(shape) >= 3:
        return dic, reorder_data(data, shape, torder)

    try:
        data = data.reshape(shape)
    except ValueError:
        warn(str(data.shape) + "cannot be shaped into" + str(shape))
        return dic, data

    return dic, data


def write_fid(filename, dic, data, torder='flat', repack=False, correct=True,
              overwrite=False):
    """
    Write a Agilent/Varian binary (fid) file.

    Parameters
    ----------
    filename : str
        Filename of Agilent/Varian binary file to write to.
    dic : dict
        Dictionary of Agilent/Varian file parameters.
    data : ndarray
        Array of NMR data.
    torder : {'f', 'r', 'o'}
        Trace ordering.  See :py:func:`read` for details.
    repack : bool, optional
        True to repack file and block headers. False leave as is.
    correct : bool, optional
        True (the default) will correct mis-sized np and nblocks values in dic.
        False will write out the incorrect values to the file header and
        record the full data set, the resulting file will not be readable by
        nmrglue.
    overwrite : bool, optional
        Set True to overwrite an existing file, False will raise a Warning if
        the file exists.

    See Also
    --------
    write_fid_lowmem : Write a Agilent/Varian binary file using mimimal
        amounts of memory
    write : Write Agilent/Varian files to a directory.

    """
    data = np.array(data)

    # convert 1D data to 2D
    if data.ndim == 1:
        data = data.reshape((1, data.shape[0]))

    # reform 3D+ data
    if data.ndim >= 3:
        data = order_data(data, torder)

    # verify data and dic shapes
    if data.shape[1] != (dic['np'] // 2):
        warn("data and np size mismatch")
        if correct:
            dic['np'] = int(data.shape[1] * 2)
    if data.shape[0] != dic["nblocks"]:
        warn("data and block size mismatch")
        if correct:
            dic['nblocks'] = int(data.shape[0])

    # open file for writing
    f = fileiobase.open_towrite(filename, overwrite=overwrite)

    if repack:
        dic = repack_fileheader(dic)

    # write the fileheader to file
    put_fileheader(f, dic2fileheader(dic))

    # determind data type
    dt = find_dtype(dic)

    if "blockheader" in dic and len(dic["blockheader"]) == data.shape[0]:
        for i in range(data.shape[0]):
            if repack:
                bh = dic2blockheader(repack_blockheader(dic["blockheader"][0]))
            else:
                bh = dic2blockheader(dic["blockheader"][i])

            trace = np.array(interleave_data(data[i]), dtype=dt)
            put_block(f, trace, dic["nbheaders"], bh)

    else:   # create a generic blockheader
        bh = dic2blockheader(make_blockheader(dic, 1))
        for i in range(data.shape[0]):
            bh[2] = int(i + 1)
            trace = np.array(interleave_data(data[i]), dtype=dt)
            put_block(f, trace, dic["nbheaders"], bh)

    f.close()
    return


def write_fid_lowmem(filename, dic, data, torder='f', repack=False,
                     overwrite=False):
    """
    Write a Agilent/Varian binary (fid) file using mimimal amounts of memory.

    File is written trace by trace with each trace read from data before
    writing to reduce memory usage.

    Parameters
    ----------
    filename : str
        Filename of Agilent/Varian binary file to write to.
    dic : dict
        Dictionary of Agilent/Varian file parameters.
    data : array_like
        Array of NMR data.
    torder : {'f', 'r', 'o'}
        Trace ordering.  See :py:func:`read` for details.
    repack : bool, optional
        True to repack file and block headers. False leave as is.
    correct : bool, optional
        True (the default) will correct mis-sized np and nblocks values in dic.
        False will write out the incorrect values to the file header and
        record the full data set, the resulting file will not be readable by
        nmrglue.
    overwrite : bool, optional
        Set True to overwrite an existing file, False will raise a Warning if
        the file exists.

    See Also
    --------
    write_fid : Write a Agilent/Varian binary.
    write : Write Agilent/Varian files to a directory.

    """
    # convert 1D data to 2D
    if data.ndim == 1:
        data = data.reshape((1, data.shape[0]))

    t2i = torder2t2i(torder)

    # verify data and dic shapes
    if data.shape[-1] != (dic["np"] // 2):
        warn("data and np size mismatch")
        if correct:
            dic['np'] = int(data.shape[1] * 2)
    nblocks = int(reduce(lambda x, y: x * y, data.shape[:-1]))
    if nblocks != dic["nblocks"]:
        warn("data and block size mismatch")
        if correct:
            dic['nblocks'] = nblocks

    # open file for writing
    f = fileiobase.open_towrite(filename, overwrite=overwrite)

    if repack:
        dic = repack_fileheader(dic)

    # write the fileheader to file
    put_fileheader(f, dic2fileheader(dic))

    # determind data type
    dt = find_dtype(dic)

    if "blockheader" in dic and len(dic["blockheader"]) == dic["nblocks"]:
        for ntrace in range(nblocks):
            if repack:
                bh = dic2blockheader(repack_blockheader(dic["blockheader"][0]))
            else:
                bh = dic2blockheader(dic["blockheader"][0])

            tup = t2i(data.shape[:-1], ntrace)
            trace = np.array(interleave_data(data[tup]), dtype=dt)
            put_block(f, trace, dic["nbheaders"], bh)

    else:   # create a generic blockheader
        bh = dic2blockheader(make_blockheader(dic, 1))
        for ntrace in range(nblocks):
            bh[2] = int(ntrace + 1)
            tup = t2i(data.shape[:-1], ntrace)
            trace = np.array(interleave_data(data[tup]), dtype=dt)
            put_block(f, trace, dic["nbheaders"], bh)
    f.close()
    return


#####################
# get/put functions #
#####################


def get_nblocks(f, nblocks, pts, nbheaders, dt, read_blockhead):
    """
    Read multiple blocks from a Agilent/Varian binary file.

    Parameters
    ----------
    f : file object
        File object of open Agilent/Varian binary file.
    nblock : int
        Number of blocks to read.
    pts : int
        Number of points per trace.
    nbheaders : int
        Number of block header in each block.
    dt : dtype
        Data type of data in binary file (real type).
    read_blockheaders : bool
        True to read the blockheader(s) and return them in the returned
        dictionary.  False ignores blockheader, reading over them.

    Returns
    -------
    dic : dict, optional
        Dictionary of blockheaders, only returned if read_blockheaders is True.
    data : 2D ndarray of size (nblocks, pts)
        Array of read blocks.

    """
    # create an empty array to hold data
    data = np.empty((nblocks, pts), dtype=dt)
    if read_blockhead:
        bdic = [0] * nblocks

    # read the data
    for i in range(nblocks):
        if read_blockhead:
            bdic[i], data[i] = get_block(f, pts, nbheaders, dt, read_blockhead)
        else:
            data[i] = get_block(f, pts, nbheaders, dt, read_blockhead)

    # return the data
    if read_blockhead:
        return bdic, data
    else:
        return data


def get_block(f, pts, nbheaders, dt, read_blockhead=False):
    """
    Read a single block from Agilent/Varian binary file.

    Parameters
    ----------
    f : file object
        File object of open Agilent/Varian binary file.
    pts : int
        Number of points per trace.
    nbheaders : int
        Number of block header in each block.
    dt : dtype
        Data type of data in binary file (real type).
    read_blockheaders : bool, optional.
        True to read the blockheader(s) and return them in the returned
        dictionary.  False (default) ignores blockheader, reading over them.

    Returns
    -------
    dic : dict, optional
        Dictionary of blockheaders, only returned if read_blockheaders is True.
    data : ndarray of size
        Array of read block.

    """
    if read_blockhead is False:  # do not return blockheaders
        for i in range(nbheaders):
            skip_blockheader(f)
        trace = get_trace(f, pts, dt)
        return trace

    else:   # read the block headers
        dic = dict()
        if nbheaders >= 1:
            dic.update(blockheader2dic(get_blockheader(f)))
        if nbheaders >= 2:
            dic["hyperhead"] = hyperheader2dic(get_hyperheader(f))
        if nbheaders >= 3:
            for i in range(2, nbheaders):
                skip_blockheader(f)
        # read the data
        trace = get_trace(f, pts, dt)

        return dic, trace


def get_nblocks_ntraces(f, nblocks, ntraces, pts, nbheaders, dt,
                        read_blockhead):
    """
    Read multiple blocks from a Agilent/Varian binary file which may have
    multiple traces per block.

    Parameters
    ----------
    f : file object
        File object of open Agilent/Varian binary file.
    nblock : int
        Number of blocks to read.
    ntrace : int
        Number of traces per block
    pts : int
        Number of points per trace.
    nbheaders : int
        Number of block header in each block.
    dt : dtype
        Data type of data in binary file (real type).
    read_blockheaders : bool
        True to read the blockheader(s) and return them in the returned
        dictionary.  False ignores blockheader, reading over them.

    Returns
    -------
    dic : dict, optional
        Dictionary of blockheaders, only returned if read_blockheaders is True.
    data : 2D ndarray of size (nblocks * ntraces, pts)
        Array of read blocks.

    """
    # create an empty array to hold data
    data = np.empty((nblocks * ntraces, pts), dtype=dt)
    if read_blockhead:
        bdic = [0] * nblocks

    # read the data
    for i in range(nblocks):
        if read_blockhead:
            bdic[i], bdata = get_block_ntraces(f, ntraces, pts, nbheaders, dt,
                                               True)
            data[i * ntraces:(i + 1) * ntraces] = bdata
        else:
            bdata = get_block_ntraces(f, ntraces, pts, nbheaders, dt, False)
            data[i * ntraces:(i + 1) * ntraces] = bdata
    if read_blockhead:
        return bdic, data
    else:
        return data


def get_block_ntraces(f, ntraces, pts, nbheaders, dt, read_blockhead=False):
    """
    Read a single block from Agilent/Varian binary file which may have
    multiple traces per block.

    Parameters
    ----------
    f : file object
        File object of open Agilent/Varian binary file.
    ntrace : int
        Number of traces per block
    pts : int
        Number of points per trace.
    nbheaders : int
        Number of block header in each block.
    dt : dtype
        Data type of data in binary file (real type).
    read_blockheaders : bool
        True to read the blockheader(s) and return them in the returned
        dictionary.  False ignores blockheader, reading over them.

    Returns
    -------
    dic : dict, optional
        Dictionary of blockheaders, only returned if read_blockheaders is True.
    data : 2D ndarray of shape (ntraces, pts)
        Array of read blocks.

    """

    if read_blockhead is False:  # do not return blockheaders
        for i in range(nbheaders):
            skip_blockheader(f)
        trace = get_trace(f, pts * ntraces, dt)
        return trace.reshape(ntraces, pts)
    else:   # read the blockheaders
        dic = dict()
        # read the headers
        if nbheaders >= 1:
            dic.update(blockheader2dic(get_blockheader(f)))
        if nbheaders >= 2:
            dic["hyperhead"] = hyperheader2dic(get_hyperheader(f))
        if nbheaders >= 3:
            for i in range(2, nbheaders):
                skip_blockheader(f)
        # read the data
        trace = get_trace(file, pts * ntraces, dt)
        return dic, trace.reshape(ntraces, pts)


def get_trace(f, pts, dt):
    """
    Read trace of pts points of dtype dt from Agilent/Varian binary file

    Endiness should be handled by dt.
    """
    bsize = pts * dt.itemsize  # number of bytes in trace
    return np.frombuffer(f.read(bsize), dt)


def get_fileheader(f):
    """
    Unpack file header parameters into a list.

    Reads the 32-byte file header from file and unpacks into a list.  Endiness
    is corrected as needed.

    Returned list contents:

    =   =========   ======================================
    N   Variable    Description
    =   =========   ======================================
    0   nblocks     data blocks in file
    1   ntraces     traces per block
    2   np          elements per trace
    3   ebytes      bytes per element
    4   tbytes      bytes per trace
    5   bbytes      bytes per block
    6   vers_id     software version, file_id status bits
    7   status      status of whole file
    8   nbheaders   number of block headers per block (1)
    =   =========   ======================================

    """
    # header is packed big-endian as 6 longs, 2 shorts, 1 long
    return struct.unpack('>6lhhl', f.read(32))


def get_blockheader(f):
    """
    Unpack block header parameters into a list.

    Reads the 28-byte block header from f and unpacks into a list.  Endiness
    is corrected as needed.

    Returned list contents:

    =   ========    ========================
    N   Variable    Description
    =   ========    ========================
    0   scale       scaling factor
    1   status      status of data in block
    2   index       block index
    3   mode        block mode
    4   ctcount     ct value of FID
    5   lpval       left phase
    6   rpval       right phase
    7   lvl         level drift correction
    8   tlt         tilt drift correction
    =   ========    ========================

    """
    # block header is packed big-endian as 4 shorts, 1 long, 4 floats
    return struct.unpack('>4hl4f', f.read(28))


def skip_blockheader(f):
    """
    Read a block header but do not unpack.

    This is a replacement for get_blockheader.  It skips f ahead 28 bytes.
    """
    f.read(28)
    return


def get_hyperheader(file):
    """
    Unpack hypercomplex header parameters to a list.

    Reads the 28-bytes block header from file and unpacks into a list. Endiness
    is corrected as needed.

    Returned list contents:

    =   ========    ================
    N   Variable    Description
    =   ========    ================
    0   s_spare1    Not Used
    1   status      status of block
    2   s_spare2    Not Used
    3   s_spart3    Not Used
    4   l_spare1    Not Used
    5   lpval1      2D left phase
    6   rpval1      2D right phase
    7   f_spare1    Not Used
    8   f_spare2    Not Used
    =   ========    ================

    """
    # hypercomplex header is packed big-endian as 4 shorts, 1 long, 4 floats
    return struct.unpack('>4hl4f', file.read(28))


def put_block(f, trace, nbheaders, bh, hh=False):
    """
    Put blockheader(s) and the trace to file.

    Parameters
    ----------
    f : file object
        Open file object to write to.
    trace : ndarray
        Trace to write to current block.
    nbheaders : int
        Number of block headers.
    bh : list
        Blockheader list.
    hh : list, optional
        Hyperheader list. Required when nbheaders >= 2.

    Notes
    -----
    When nbheaders > 2, additional headers are written as all zeros.

    """
    # write the block headers
    if nbheaders >= 1:
        put_blockheader(f, bh)

    if nbheaders >= 2:
        if hh is False:
            raise Exception("Hyperheader required")
        put_hyperheader(f, hh)

    # write any additional blockheaders as 0s if needed
    for i in range(nbheaders - 2):
        put_blockheader(f, [0] * 9)

    # write the trace
    put_trace(f, trace)

    return


def put_trace(f, trace):
    """
    Write a trace to file f.
    """
    f.write(trace.tostring())
    return


def put_fileheader(f, fh):
    """
    Write a fileheader list to file (32-bytes written).

    Parameters
    ----------
    f : file object
        Open file object to write to.
    fh : list with 9 elements
        Fileheader list.

    """
    f.write(struct.pack('>6lhhl', *fh))
    return


def put_blockheader(f, bh):
    """
    Write a blockheader list to file (28-bytes written)

    Parameters
    ----------
    f : file object
        Open file object to write to.
    bh : list with 9 elements
        Blockheaders list.

    """
    f.write(struct.pack('>4hl4f', *bh))
    return


def put_hyperheader(f, hh):
    """
    Write hyperheader list to file (28-bytes written)

    Parameters
    ----------
    f : file object
        Open file object to write to.
    hh : list with 9 elements
        Hyperheader list.

    """
    f.write(struct.pack('>4hl4f', *hh))
    return


#########################
# dictionary conversion #
#########################


def hyperheader2dic(head):
    """
    Convert a hypercomplex block header into a Python dictionary.
    """
    dic = dict()
    dic["s_spare1"] = head[0]
    dic["status"] = head[1]
    dic["s_spare2"] = head[2]
    dic["s_spare3"] = head[3]
    dic["l_spare1"] = head[4]
    dic["lpval1"] = head[5]
    dic["rpval1"] = head[6]
    dic["f_spare1"] = head[7]
    dic["f_spare2"] = head[8]

    # unpack the status bits
    dic["UHYPERCOMPLEX"] = (dic["status"] & 0x2) // 0x2

    return dic


def repack_hyperheader(dic):
    """
    Repack a hyperheader dictionary bit flag parameters into status.
    """
    dic["status"] = dic["UHYPERCOMPLEX"] * 0x2
    return dic


def dic2hyperheader(dic):
    """
    Convert a Python dictionary into a hypercomplex block header list.

    Does not repack status from bit flags.
    """
    head = [0] * 9
    head[0] = dic["s_spare1"]
    head[1] = dic["status"]
    head[2] = dic["s_spare2"]
    head[3] = dic["s_spare3"]
    head[4] = dic["l_spare1"]
    head[5] = dic["lpval1"]
    head[6] = dic["rpval1"]
    head[7] = dic["f_spare1"]
    head[8] = dic["f_spare2"]
    return head


def make_blockheader(filedic=False, index=1):
    """
    Make a generic blockheader dictionary with a given block index.

    filedic can be provided for status flags, if not provided creates
    header for float32 data

    """
    dic = dict()

    if filedic is False:
        # common status flags
        dic["S_DATA"] = filedic["S_DATA"]
        dic["S_SPEC"] = filedic["S_SPEC"]
        dic["S_32"] = filedic["S_32"]
        dic["S_FLOAT"] = filedic["S_FLOAT"]
        dic["S_COMPLEX"] = filedic["S_COMPLEX"]
        dic["S_HYPERCOMPLEX"] = filedic["S_HYPERCOMPLEX"]
    else:
        # common status flags
        dic["S_DATA"] = 1
        dic["S_SPEC"] = 0
        dic["S_32"] = 0
        dic["S_FLOAT"] = 1
        dic["S_COMPLEX"] = 0
        dic["S_HYPERCOMPLEX"] = 0

    # blockheader specific status flags
    dic["MORE_BLOCKS"] = 1
    dic["NP_CMPLX"] = 0
    dic["NF_CMPLX"] = 0
    dic["NI_CMPLX"] = 0
    dic["NI2_CMPLX"] = 0

    # mode flags
    dic["NP_PHMODE"] = 0
    dic["NP_AVMODE"] = 0
    dic["NP_PWRMODE"] = 0
    dic["NF_PHMODE"] = 0
    dic["NF_AVMODE"] = 0
    dic["NF_PWRMODE"] = 0
    dic["NI_PHMODE"] = 0
    dic["NI_AVMODE"] = 0
    dic["NI_PWRMODE"] = 0
    dic["NI2_PHMODE"] = 0
    dic["NI2_AVMODE"] = 0
    dic["NI2_PWRMODE"] = 0

    # main parameters
    dic["scale"] = 0
    dic["status"] = 0
    dic["index"] = index
    dic["mode"] = 0
    dic["ctcount"] = 1
    dic["lpval"] = 0.0
    dic["rpval"] = 0.0
    dic["lvl"] = 0.0
    dic["tlt"] = 0.0

    repack_blockheader(dic)  # repack the header
    return dic


def blockheader2dic(head):
    """
    Convert a block header list into a Python dictionary.
    """
    dic = dict()

    dic["scale"] = head[0]
    dic["status"] = head[1]
    dic["index"] = head[2]
    dic["mode"] = head[3]
    dic["ctcount"] = head[4]
    dic["lpval"] = head[5]
    dic["rpval"] = head[6]
    dic["lvl"] = head[7]
    dic["tlt"] = head[8]

    # unpack the status parameters
    dic["S_DATA"] = (dic["status"] & 0x1) // 0x1
    dic["S_SPEC"] = (dic["status"] & 0x2) // 0x2
    dic["S_32"] = (dic["status"] & 0x4) // 0x4
    dic["S_FLOAT"] = (dic["status"] & 0x8) // 0x8
    dic["S_COMPLEX"] = (dic["status"] & 0x10) // 0x10
    dic["S_HYPERCOMPLEX"] = (dic["status"] & 0x20) // 0x20

    dic["MORE_BLOCKS"] = (dic["status"] & 0x80) // 0x80
    dic["NP_CMPLX"] = (dic["status"] & 0x100) // 0x100
    dic["NF_CMPLX"] = (dic["status"] & 0x200) // 0x200
    dic["NI_CMPLX"] = (dic["status"] & 0x400) // 0x400
    dic["NI2_CMPLX"] = (dic["status"] & 0x800) // 0x800

    # unpack the mode parameter
    dic["NP_PHMODE"] = (dic["mode"] & 0x1) // 0x1
    dic["NP_AVMODE"] = (dic["mode"] & 0x2) // 0x2
    dic["NP_PWRMODE"] = (dic["mode"] & 0x4) // 0x4

    dic["NF_PHMODE"] = (dic["mode"] & 0x10) // 0x10
    dic["NF_AVMODE"] = (dic["mode"] & 0x20) // 0x20
    dic["NF_PWRMODE"] = (dic["mode"] & 0x40) // 0x40

    dic["NI_PHMODE"] = (dic["mode"] & 0x100) // 0x100
    dic["NI_AVMODE"] = (dic["mode"] & 0x200) // 0x200
    dic["NI_PWRMODE"] = (dic["mode"] & 0x400) // 0x400

    dic["NI2_PHMODE"] = (dic["mode"] & 0x1000) // 0x1000
    dic["NI2_AVMODE"] = (dic["mode"] & 0x2000) // 0x2000
    dic["NI2_PWRMODE"] = (dic["mode"] & 0x4000) // 0x4000

    return dic


def repack_blockheader(dic):
    """
    Repack blockheader dic bit flag parameters into status and mode.
    """

    dic["status"] = (dic["S_DATA"] * 0x1 + dic["S_SPEC"] * 0x2 +
                     dic["S_32"] * 0x4 + dic["S_FLOAT"] * 0x8 +
                     dic["S_COMPLEX"] * 0x10 + dic["S_HYPERCOMPLEX"] * 0x20 +
                     dic["MORE_BLOCKS"] * 0x80 + dic["NP_CMPLX"] * 0x100 +
                     dic["NF_CMPLX"] * 0x200 + dic["NI_CMPLX"] * 0x400 +
                     dic["NI2_CMPLX"] * 0x800)

    dic["mode"] = (dic["NP_PHMODE"] * 0x1 + dic["NP_AVMODE"] * 0x2 +
                   dic["NP_PWRMODE"] * 0x4 + dic["NF_PHMODE"] * 0x10 +
                   dic["NF_AVMODE"] * 0x20 + dic["NF_PWRMODE"] * 0x40 +
                   dic["NI_PHMODE"] * 0x100 + dic["NI_AVMODE"] * 0x200 +
                   dic["NI_PWRMODE"] * 0x400 + dic["NI2_PHMODE"] * 0x1000 +
                   dic["NI2_AVMODE"] * 0x2000 + dic["NI2_PWRMODE"] * 0x4000)
    return dic


def dic2blockheader(dic):
    """
    Convert a python dictionary into block header list.

    Does not repack status and mode from bit flags.
    """
    head = [0] * 9
    head[0] = dic["scale"]
    head[1] = dic["status"]
    head[2] = dic["index"]
    head[3] = dic["mode"]
    head[4] = dic["ctcount"]
    head[5] = dic["lpval"]
    head[6] = dic["rpval"]
    head[7] = dic["lvl"]
    head[8] = dic["tlt"]
    return head


def fileheader2dic(head):
    """
    Convert fileheader list into a Python dictionary
    """
    dic = dict()
    dic["nblocks"] = head[0]
    dic["ntraces"] = head[1]
    dic["np"] = head[2]
    dic["ebytes"] = head[3]
    dic["tbytes"] = head[4]
    dic["bbytes"] = head[5]
    dic["vers_id"] = head[6]
    dic["status"] = head[7]
    dic["nbheaders"] = head[8]

    # unpack the status parameter
    dic["S_DATA"] = (dic["status"] & 0x1) // 0x1
    dic["S_SPEC"] = (dic["status"] & 0x2) // 0x2
    dic["S_32"] = (dic["status"] & 0x4) // 0x4
    dic["S_FLOAT"] = (dic["status"] & 0x8) // 0x8
    dic["S_COMPLEX"] = (dic["status"] & 0x10) // 0x10
    dic["S_HYPERCOMPLEX"] = (dic["status"] & 0x20) // 0x20
    dic["S_ACQPAR"] = (dic["status"] & 0x80) // 0x80
    dic["S_SECND"] = (dic["status"] & 0x100) // 0x100
    dic["S_TRANSF"] = (dic["status"] & 0x200) // 0x200
    dic["S_NP"] = (dic["status"] & 0x800) // 0x800
    dic["S_NF"] = (dic["status"] & 0x1000) // 0x1000
    dic["S_NI"] = (dic["status"] & 0x2000) // 0x2000
    dic["S_NI2"] = (dic["status"] & 0x4000) // 0x4000

    return dic


def repack_fileheader(dic):
    """
    Repack blockheader dic bit flag parameters into status and mode.
    """
    dic["status"] = (dic["S_DATA"] * 0x1 + dic["S_SPEC"] * 0x2 +
                     dic["S_32"] * 0x4 + dic["S_FLOAT"] * 0x8 +
                     dic["S_COMPLEX"] * 0x10 + dic["S_HYPERCOMPLEX"] * 0x20 +
                     dic["S_ACQPAR"] * 0x80 + dic["S_SECND"] * 0x100 +
                     dic["S_TRANSF"] * 0x200 + dic["S_NP"] * 0x800 +
                     dic["S_NF"] * 0x1000 + dic["S_NI"] * 0x2000 +
                     dic["S_NI2"] * 0x4000)
    return dic


def dic2fileheader(dic):
    """
    Convert a Python dictionary into a fileheader list

    Does not repack status from bit flags
    """

    head = [0] * 9
    head[0] = dic["nblocks"]
    head[1] = dic["ntraces"]
    head[2] = dic["np"]
    head[3] = dic["ebytes"]
    head[4] = dic["tbytes"]
    head[5] = dic["bbytes"]
    head[6] = dic["vers_id"]
    head[7] = dic["status"]
    head[8] = dic["nbheaders"]

    return head

##################
# misc functions #
##################


def find_shape(pdic):
    """
    Determine the shape of a Agilent/Varian file from the procpar dictionary
    """
    # Varian files are typically either from imaging experiments or
    # from NMR experiments.  For both the direct dimension (R+I) is stored in
    # the np parameters in the procpar file.  In addition an array may be
    # present.

    # Imaging experiments have a 'seqcon' parameter present in the procpar and
    # have indirect dimension are set by nv, nv2, and nv3.

    # NMR experiments has indirect dimension shapes (R|I) are stored in
    # ni,ni2,ni3 with the phase arrays stored as phase,phase2,phase3

    # Thanks to the VeSPA project (http://scion.duhs.duke.edu/vespa/) for
    # information on imaging experiments

    try:
        shape = [int(pdic["np"]["values"][0]) // 2]
    except:
        # when shape finding fails issue warning and return None
        warn("shape cannot be determined.")
        return None

    # check for an array, we will deal only with the inner most array.
    # When multiple parameters are arrayed the ordering of indirect dimension
    # must also be known, which typically is not provided, so we will leave
    # this edge case to the user.
    if "array" in pdic:
        array_name = pdic['array']['values'][0]
        array_name = array_name.split(',')[-1]   # keep only innermost array
        if array_name.startswith('phase') is False and array_name in pdic:
            shape.insert(0, len(pdic[array_name]["values"]))

    # imaging files have a seqcon parameter
    if 'seqcon' in pdic:
        if "nv" in pdic:
            s = max(int(pdic["nv"]["values"][0]), 1)
            if s > 1:
                shape.insert(0, s)
        if "nv2" in pdic:
            s = max(int(pdic["nv2"]["values"][0]), 1)
            if s > 1:
                shape.insert(0, s)

        if "nv3" in pdic:
            s = max(int(pdic["nv3"]["values"][0]), 1)
            if s > 1:
                shape.insert(0, s)
        return tuple(shape)

    # assume we have NMR data
    if "ni" in pdic:
        multi = 2       # assume R+I in cases where no phase parameter.
        if "phase" in pdic:
            multi = len(pdic["phase"]["values"])
        s = max(int(pdic["ni"]["values"][0]), 1)
        shape.insert(0, s * multi)

    if "ni2" in pdic:
        multi = 2
        if "phase2" in pdic:
            multi = len(pdic["phase2"]["values"])
        s = max(int(pdic["ni2"]["values"][0]), 1)
        shape.insert(0, s * multi)

    if "ni3" in pdic:
        multi = 2
        if "phase3" in pdic:
            multi = len(pdic["phase"]["values"])
        s = max(int(pdic["ni3"]["values"][0]), 1)
        shape.insert(0, s * multi)

    return tuple(shape)


def find_cdtype(dic):
    """
    Find the complex dtype from a Agilent/Varian dictionary
    """
    if dic["S_FLOAT"] == 1:
        return np.dtype("complex64")
    else:
        if dic["S_32"] == 1:
            return np.dtype("complex128")
        else:
            return np.dtype("complex64")


def find_dtype(dic):
    """
    Find the real dtype from a dictionary
    """
    if dic["S_FLOAT"] == 1:
        return np.dtype('>f4')  # float32
    else:
        if dic["S_32"] == 1:
            return np.dtype('>i4')  # int32
        else:
            return np.dtype('>i2')  # int16


def uninterleave_data(data):
    """
    Unpack interleaved real, imag data

    ==========  ============
    data dtype  Return dtype
    ==========  ============
    int16       'complex64'
    float32     'complex64'
    int32       'complex128'
    ==========  ============

    """
    # determind the output dtype
    rdt = data.dtype.name

    if rdt == 'int16' or rdt == "float32":
        cdt = "complex64"
    elif rdt == 'int32':
        cdt = "complex128"
    else:
        raise ValueError("unknown dtype")

    return data[..., ::2] + np.array(data[..., 1::2] * 1.j, dtype=cdt)


def interleave_data(data_in):
    """
    Interleave real, imag data

    Does not check if resulting dtype is a valid Agilent/Varian dtype
    """
    size = list(data_in.shape)
    size[-1] = size[-1] * 2
    data_out = np.empty(size, dtype=data_in.real.dtype)
    data_out[..., ::2] = data_in.real
    data_out[..., 1::2] = data_in.imag
    return data_out

###########################
# procpar reading/writing #
###########################

# procpar functions


def read_procpar(filename):
    """
    Read a procpar file returning a dictionary of procpar parameters.
    """
    with open(filename, 'rb') as f:
        dic = dict()
        length = os.stat(filename).st_size

        # test to see if end of file
        while f.tell() != length:
            p = get_parameter(f)
            dic[p["name"]] = p
        return dic


def get_parameter(f):
    """
    Reads a procpar parameter from a file object.

    Returns a dictionary with the attributes of the parameter.

    """
    dic = dict()

    # read and decode the first line
    line = f.readline().decode().split()

    dic["name"] = line[0]
    dic["subtype"] = line[1]
    dic["basictype"] = line[2]
    dic["maxvalue"] = line[3]
    dic["minvalue"] = line[4]
    dic["stepsize"] = line[5]
    dic["Ggroup"] = line[6]
    dic["Dgroup"] = line[7]
    dic["protection"] = line[8]
    dic["active"] = line[9]
    dic["intptr"] = line[10]

    # read in the values of the parameter
    line = f.readline().decode()
    num = int(line.split()[0])
    values = []

    if dic["basictype"] == "1":     # real values, only one line
        values = line.split()[1:]
    elif dic["basictype"] == "2":   # strings, may have multiple lines
        values.append(line.split("\"")[1])  # split on "s
        for i in range(num - 1):
            values.append(f.readline().decode().split("\"")[1])

    dic["values"] = values
    line = f.readline().decode()

    # read and decode the enumerables
    dic["enumerable"] = line.split()[0]

    if dic["enumerable"] != "0":
        if dic["basictype"] == "1":     # reals
            dic["enumerables"] = line.split()[1:]
        elif dic["basictype"] == "2":   # strings
            dic["enumerables"] = line.split("\"")[1::2]

    return dic


def write_procpar(filename, dic, overwrite=False):
    """
    Write a Agilent/Varian procpar file from a dictionary
    """
    # open the file for writing
    f = fileiobase.open_towrite(filename, overwrite=overwrite, mode='w')

    for key in dic.keys():  # loop over the parameters

        d = dic[key]
        # print(out the first line)
        print(d["name"], d["subtype"], d["basictype"],
              d["maxvalue"], d["minvalue"], d["stepsize"],
              d["Ggroup"], d["Dgroup"], d["protection"],
              d["active"], d["intptr"], file=f)

        # print out the next line(s) (and more if strings)
        if d["basictype"] == "1":   # real values, one line
            print(len(d["values"]), end=' ', file=f)  # don't end the line
            for value in d["values"]:
                print(value, end=' ', file=f)    # still not yet
            print("", file=f)   # now end the line

        elif d["basictype"] == "2":     # strings, may have multiple lines
            print(len(d["values"]), end=' ', file=f)  # don't end the line
            for value in d["values"]:
                # now end the line (for each string)
                print('"' + value + '"', file=f)

        # print out the last line
        print(d["enumerable"], end=' ', file=f)

        if d["enumerable"] != "0":
            for e in d["enumerables"]:
                if d["basictype"] == "1":  # reals
                    print(e, end=' ', file=f)
                elif d["basictype"] == "2":  # strings
                    print('"' + e + '"', end=' ', file=f)
        print("", file=f)   # end the enumerable line

    f.close()

    return

subtypes = ["undefined", "real", "string", "delay", "flag", "frequency",
            "pulse", "integer"]

basictypes = ["undefined", "real", "string"]

############################################
# low memory numpy.ndarray emulating class #
############################################


class fid_nd(fileiobase.data_nd):
    """
    Emulate a ndarray objects without loading data into memory for low memory
    reading of Agilent/Varian fid files which must have one trace per block.

    * slicing operations return ndarray objects.
    * can iterate over with expected results.
    * transpose and swapaxes methods create a new objects with correct axes
      ordering.
    * has ndim, shape, and dtype attributes.

    Parameters
    ----------
    filename : str
        Filename of Agilent/Varian binary file.
    i2t_func : function
        Python function which maps an index to a trace.
    fshape : tuple of ints, optional
        Shape of data in file, if None will be assumed to be 2D data.
    order : tuple
        Ordering of axes compared to file. None will results in (0, 1, 2, ...)
        ordering.

    """
    def __init__(self, filename, i2t_func, fshape=None, order=None):
        """
        Create and set up object.
        """
        # read the file dictionary
        f = open(filename, 'rb')
        dic = fileheader2dic(get_fileheader(f))
        f.close()

        # check fshape
        if fshape is None:
            # by default open as 2D with nblocks,np/2 fshape
            fshape = (dic['nblocks'], int(dic["np"] / 2))
        else:
            # check that last dimension has np/2 points
            if fshape[-1] != int(dic["np"] / 2):
                s = "last dimension should have size %i" % (int(dic["np"] / 2))
                raise ValueError(s)
            # product of all but last dim should be number of blocks
            if reduce(lambda x, y: x * y, fshape[:-1]) != dic['nblocks']:
                s = "number of traces in file does not match fshape"
                raise ValueError(s)

        # check order
        if order is None:
            order = range(len(fshape))

        # finalize
        self.fdtype = find_dtype(dic)
        self.pts = dic["np"]
        self.nbh = dic["nbheaders"]
        self.bbytes = dic["bbytes"]
        self.dtype = find_cdtype(dic)
        self.filename = filename
        self.i2t = i2t_func
        self.fshape = fshape
        self.order = order
        self.__setdimandshape__()   # set ndim and shape attributes

    def __fcopy__(self, order):
        """
        Create a copy.
        """
        n = fid_nd(self.filename, self.i2t, self.fshape, order)
        return n

    def __fgetitem__(self, slices):
        """
        Return ndarray of selected values

        slices is a well formatted tuple of slices
        """
        # seperate the last slice from the first slices
        lslice = slices[-1]
        fslice = slices[:-1]

        # and the same for fshape
        lfshape = self.fshape[-1]
        ffshape = self.fshape[:-1]

        # find the output size and make a in/out nd interator
        osize, nd_iter = fileiobase.size_and_ndtofrom_iter(ffshape, fslice)
        osize.append(len(range(lfshape)[lslice]))

        # create an empty array to store the selected slices
        out = np.empty(tuple(osize), dtype=self.dtype)

        with open(self.filename, 'rb') as f:

            # read in the data trace by trace
            for out_index, in_index in nd_iter:

                # determine the trace number from the index
                ntrace = self.i2t(ffshape, in_index)

                # seek to the correct place in the file
                f.seek(ntrace * self.bbytes + 32)

                # retrive trace and save to output
                trace = get_block(f, self.pts, self.nbh, self.fdtype, False)
                trace = uninterleave_data(trace)
                out[out_index] = trace[lslice]

        return out



"""
nmrglue table functions.

nmrglue uses numpy records array as stores of various data (peak tables,
trajectories, etc).  This module provides functions to read and write records
arrays from disk.  Formatting of the numeric values is left to Python's str
function and only the data type need be specified.  In addition this module
contains functions to convert nmrglue's table format NMRPipe's table format.
"""

import numpy as np
from . import fileiobase


def pipe2glue(pcomments, pformat, rec):
    """
    Convert a NMRPipe table to a nmrglue table

    Parameters
    ----------
    pcomments : list
        List of NMRPipe comment lines.
    pformats : list
        List of NMRPipe table column formats strings.
    rec : recarray
        Records array with named fields.

    Returns
    -------
    comments : list
        List of comments
    rec : recarray
        Records array with named fields.

    """
    # add a "#" to the list of comments and we are done
    comments = ["# " + c for c in pcomments]
    return comments, rec


def glue2pipe(comments, rec):
    """
    Convert a nmrglue table to a NMRPipe table.

    Parameters
    ----------
    comments : list
        List of comments
    rec : recarray
        Records array with named fields.

    Returns
    -------
    pcomments : list
        List of NMRPipe comment lines.
    pformats : list
        List of NMRPipe table column formats strings.  This list is guessed
        from the data types and precision in the reconrds array.  This may not
        be the exact format desired, edit this to your liking.
    rec : recarray
        Records array with named fields.

    """
    # add REMARK to each comment
    pcomments = ["REMARK " + c for c in comments]

    # guess the pipe format strings
    pformat = [guess_pformat(rec[t]) for t in rec.dtype.names]

    return pcomments, pformat, rec


def guess_pformat(col):
    """
    Guess a NMRPipe table column format string given a column.

    Parameters
    ----------
    col : ndarray
        Array from a records array.

    Returns
    -------
    s : str
        String for formatting NMRPipe table.

    """
    kind = col.dtype.kind

    if kind == 'S' or kind == 'a':  # string
        return '%s'

    if kind == 'i' or kind == 'u':  # integer (signed or unsigned)
        # N is the number of digits in largest value, or 1
        N = max(np.ceil(np.log(np.abs(col).max()) / np.log(10)), 1)
        # +1 for sign
        return '%{0}d'.format(int(N + 1))

    if kind == 'f':
        # will be either %+e or %N.3f, see if 'e' is in %g to determine
        if True in ['e' in '%g' % (v) for v in col]:
            return '%+e'
        else:
            N = max(np.ceil(np.log(np.abs(col).max()) / np.log(10)), 1)
            # +1 for sign, +1 for decimal points, +3 for precision
            return '%{0}.3f'.format(int(N + 5))

    # remaining kinds: 'c' - complex, 'b' - boolean, 'U' - unicode, 'V' - void
    raise ValueError("unknown kind %s in column" % (kind))


def read(filename):
    """
    Read a nmrglue table file.

    Parameters
    ----------
    filename : str
        Filename of nmrglue table file to read.

    Returns
    -------
    comments : list
        List of comments (strings terminated with newline)
    rec : recarray
        Records array with named fields.

    """
    # pull out the comment lines from the file (start with #)
    f = open(filename, 'rb')
    comments = [l for l in f if l[0] == '#']
    f.close()

    # find the line beginning with # NAMES and parse out the column names
    nl = [i for i, l in enumerate(comments) if l[:7] == "# NAMES"]
    if len(nl) != 1:
        raise IOError("%s does not have a # NAMES line" % (filename))
    dtd = {'names': comments.pop(nl[0])[7:].split()}

    # find the line beginning with # DTYPE and parse out the column names
    dl = [i for i, l in enumerate(comments) if l[:9] == "# FORMATS"]
    if len(dl) != 1:
        raise IOError("%s does not have a # FORMATS line" % (filename))
    dtd['formats'] = comments.pop(dl[0])[9:].split()

    # return the data as a records array
    return comments, np.atleast_1d(np.recfromtxt(filename, dtype=dtd))


def write(filename, comments, rec, overwrite=False):
    """
    Write a nmrglue table to file.

    Parameters
    ----------
    filename : str
        Filename of file to write table to.
    comments : list
        List of comments (strings terminated with newline).
    rec : recarray
        Records array to write to file.
    overwrite : bool, optional
        True to overwrite file if it exists. False will raise an Warning if the
        file exists.

    """
    # open the file for writing
    f = fileiobase.open_towrite(filename, overwrite)

    # write out the comment lines at the top of the file
    for c in comments:
        f.write(c)

    # Determind the list of column names
    names = rec.dtype.names

    # Determind the list of column formats
    sizes = [rec.dtype[n].itemsize for n in names]
    kinds = [rec.dtype[n].kind for n in names]
    formats = [k + str(s) for s, k in zip(sizes, kinds)]

    # write out the NAMES and FORMATS lines
    f.write("# NAMES " + " ".join(names) + "\n")
    f.write("# FORMATS " + " ".join(formats) + "\n")

    # maximum string length for each column
    col_len = [max([len(str(i)) for i in rec[n]]) for n in names]

    # write out each line of the table
    for row in rec:
        s = " ".join([str(v).ljust(l) for v, l in zip(row, col_len)])
        f.write(s + "\n")
    f.close()


# Row functions
def insert_row(rec, N, row):
    """
    Insert a row into a records array before row number N.

    Parameters
    ----------
    rec : recarray
        Records array.
    N : int
        Row number to insert new row before.
    row : array_like
        Array or similar object which will be converted into a new row.

    Returns
    -------
    new_rec : recarray
        New records array with inserted row.

    """
    return np.insert(rec, N, row)


def append_row(rec, row):
    """
    Append a row to the end of a records array.

    Parameters
    ----------
    rec : recarray
        Records array.
    row : array_like
         Array or similar object which will be converted into a new row.

    Returns
    -------
    new_rec : recarray
        New records array with inserted row.

    """
    N = len(rec)
    return insert_row(rec, N, row)


def delete_row(rec, N):
    """
    Delete a row from a records array.

    Parameters
    ----------
    rec : recarray
        Records array.
    N : int
        Row number to delete.

    Returns
    -------
    new_rec : recarray
        New records array with row deleted.

    See Also
    --------
    reorder_rows : delete multiple rows in a single call.

    """
    return np.delete(rec, N)


def reorder_rows(rec, new_order):
    """
    Reorder or delete rows in a records array.

    This function can also be used to delete multiple rows from a records
    array, only the rows in the new_order list are retained in the new records
    array.

    Parameters
    ----------
    rec : recarray
        Records array.
    new_order : list
        List of row indices and order in new records array.  Only the rows in
        this list are retained in the new records array.  Therefore this
        function can also be used to delete multiple rows from a records
        array.

    Returns
    -------
    new_rec : recarray
        New records array with rows reordered.

    """
    return np.take(rec, new_order)


# Column functions
def append_column(rec, col, name=None, format=None):
    """
    Append a column to the end of a records array.

    Parameters
    ----------
    rec : recarray
        Records array.
    col : array_like
        Array or similar object which will be converted into the new column.
    name : str, optional
        Name of the column. If None col.dtypes.name will be used.
    format : dtype, optional
        Data type to convert the new column into before appending. Required if
        col is not an ndarray.

    Returns
    -------
    new_rec : recarray
        New records array with column appended.

    """
    N = len(rec.dtype.descr)
    return insert_column(rec, N, col, name, format)


def insert_column(rec, N, col, name=None, format=None):
    """
    Insert a column into a records array.

    Parameters
    ----------
    rec : recarray
        Records array.
    col : array_like
        Array or similar object which will be converted into the new column.
    N : int
        Column number to insert new column before.
    name : str, optional
        Name of the column. If None col.dtypes.name will be used.
    format : dtype, optional
        Data type to convert the new column into before appending. Required if
        col in not an ndarray.

    Returns
    -------
    new_rec : recarray
        New records array with column inserted.

    """
    col = np.array(col)

    # get name and format parameter from column if not provided
    if name is None:
        if col.dtype.names is not None:
            name = col.dtype.names
        else:
            raise ValueError("Must provide a name for the column")
    if format is None:
        format = col.dtype.str

    # insert the new column (name, format) to the table dtypes
    dtd = rec.dtype.descr
    dtd.insert(N, (name, format))

    # create the new table with an additional column
    new_rec = np.empty(rec.shape, dtd)

    # fill in the old columns
    for n in rec.dtype.names:
        new_rec[n] = rec[n]
    # and the new column
    new_rec[name] = col.astype(format)

    return np.atleast_1d(np.rec.array(new_rec))


def delete_column(rec, N):
    """
    Delete a column from a records array.

    Parameters
    ----------
    rec : recarray
        Records array.
    N : int
        Column number to delete.

    Returns
    -------
    new_rec : recarray
        New records array with column deleted.

    See Also
    --------
    reorder_columns : Delete multiple columns from a records array.

    """
    # remove the column from the list of columns.
    dtd = rec.dtype.descr
    dtd.pop(N)

    # create the new records array and fill it in
    new_rec = np.empty(rec.shape, dtd)
    for n in new_rec.dtype.names:
        new_rec[n] = rec[n]
    return np.atleast_1d(np.rec.array(new_rec))


def reorder_columns(rec, new_order):
    """
    Reorder or delete columns in a records array.

    Parameters
    ----------
    rec : recarray
        Records array.
    new_order : list
        List of column indices and order in new records array.  Only the
        columns in this list are retained in the new records array.
        Therefore this function can also be used to delete multiple columns
        from a records array.

    Returns
    -------
    new_rec : recarray
        New records array with columns reordered.

    """
    # reorder the dtype description list
    dtd = rec.dtype.descr
    new_dtd = [dtd[i] for i in new_order]

    # create the new array and fill it in.
    new_rec = np.empty(rec.shape, new_dtd)
    for n in new_rec.dtype.names:
        new_rec[n] = rec[n]
    return np.atleast_1d(np.rec.array(new_rec))



""" Unit tests for nmrglue/fileio/pipe.py module """

import os
import tempfile
import glob

from numpy.testing import assert_array_equal
import nmrglue as ng

# NMRPipe files being tested, these are created by the script:
# create_test_data_nmrpipe.sh
# they are checked into the repository, so NMRPipe is not required to
# run the tests
DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')

NMRPIPE_1D_TIME = os.path.join(DATA_DIR, 'nmrpipe_1d_time.fid')
NMRPIPE_1D_FREQ = os.path.join(DATA_DIR, 'nmrpipe_1d_freq.fid')
NMRPIPE_1D_EXT = os.path.join(DATA_DIR, 'nmrpipe_1d_ext.fid')

NMRPIPE_2D_TIME = os.path.join(DATA_DIR, 'nmrpipe_2d_time.fid')
NMRPIPE_2D_FREQ = os.path.join(DATA_DIR, 'nmrpipe_2d_freq.ft2')
NMRPIPE_2D_TIME_TP = os.path.join(DATA_DIR, 'nmrpipe_2d_time_tp.fid')
NMRPIPE_2D_FREQ_TP = os.path.join(DATA_DIR, 'nmrpipe_2d_freq_tp.ft2')

NMRPIPE_3D_TIME = os.path.join(DATA_DIR, 'nmrpipe_3d_time.dir',
                               'nmrpipe_3d_time_%03d.fid')
NMRPIPE_3D_FREQ = os.path.join(DATA_DIR, 'nmrpipe_3d_freq.dir',
                               'nmrpipe_3d_freq_%03d.ft3')
NMRPIPE_3D_TIME_STREAM = os.path.join(DATA_DIR, 'nmrpipe_3d_time.fid')
NMRPIPE_3D_FREQ_STREAM = os.path.join(DATA_DIR, 'nmrpipe_3d_freq.ft3')

NMRPIPE_4D_TIME_2 = os.path.join(DATA_DIR, 'nmrpipe_4d_time_2.dir',
                                 'nmrpipe_4d_time_%03d_%03d.fid')
NMRPIPE_4D_TIME_1 = os.path.join(DATA_DIR, 'nmrpipe_4d_time_1.dir',
                                 'nmrpipe_4d_time_%03d.fid')
NMRPIPE_4D_TIME_STREAM = os.path.join(DATA_DIR, 'nmrpipe_4d_time.fid')

NMRPIPE_4D_FREQ_1 = os.path.join(DATA_DIR, 'nmrpipe_4d_freq_1.dir',
                                 'nmrpipe_4d_freq_%03d.ft4')
NMRPIPE_4D_FREQ_2 = os.path.join(DATA_DIR, 'nmrpipe_4d_freq_2.dir',
                                 'nmrpipe_4d_freq_%03d_%03d.ft4')
NMRPIPE_4D_FREQ_STREAM = os.path.join(DATA_DIR, 'nmrpipe_4d_freq.ft4')
NMRPIPE_TABLE = os.path.join(DATA_DIR, 'test.tab')


def check_simple_roundtrip(dic, data, n_percents=0, lowmem=False):
    """ Check write/read roundtrip for simple data. """
    try:
        # create one or more temporary file name(s)
        base_fname = tempfile.mktemp(prefix='nmrgluetest', dir='.')
        tmpfname = base_fname + '_%03d' * n_percents

        # write data to it, read back and check that same
        if lowmem:
            ng.pipe.write_lowmem(tmpfname, dic, data)
            rdic, rdata = ng.pipe.read_lowmem(tmpfname)
            s = tuple([0] * data.ndim)
            assert data[s] == rdata[s]
        else:
            ng.pipe.write(tmpfname, dic, data)
            rdic, rdata = ng.pipe.read(tmpfname)
            assert_array_equal(data, rdata)
        assert dic == rdic

    finally:    # remove the temporary files
        for f in glob.glob(base_fname + "*"):
            os.remove(f)


def check_ppm_limits(dic, data, dim, limits):
    """ Check PPM Limits """
    uc0 = ng.pipe.make_uc(dic, data, dim=dim)
    l0, l1 = uc0.ppm_limits()
    assert round(l0, 2) == limits[0]
    assert round(l1, 2) == limits[1]


def test_1d_time():
    """ read/write of 1D NMRPipe time domain file. """
    dic, data = ng.pipe.read(NMRPIPE_1D_TIME)
    assert data.shape == (16, )
    assert data.dtype == 'complex64'
    assert data[0].real == 1.
    assert data[0].imag == -1.
    assert data[1].real == 2.
    assert data[1].imag == -2.
    check_simple_roundtrip(dic, data)


def test_1d_freq():
    """ read/write of 1D NMRPipe frequency domain file. """
    dic, data = ng.pipe.read(NMRPIPE_1D_FREQ)
    assert data.shape == (16, )
    assert data.dtype == 'float32'
    assert data[0].real == 1.
    assert data[1].real == 2.
    check_simple_roundtrip(dic, data)
    check_ppm_limits(dic, data, 0, [149.00, 55.25])


def test_1d_ext():
    """ read/write of 1D NMRPipe file with an EXTacted region. """
    dic, data = ng.pipe.read(NMRPIPE_1D_EXT)
    assert data.shape == (8, )
    assert data.dtype == 'float32'
    assert data[0].real == 1.
    assert data[1].real == 2.
    check_simple_roundtrip(dic, data)
    check_ppm_limits(dic, data, 0, [130.25, 86.50])


def test_2d_time():
    """ read/write of 2D NMRPipe time domain file. """
    dic, data = ng.pipe.read(NMRPIPE_2D_TIME)
    assert data.shape == (4, 8)
    assert data.dtype == 'complex64'
    assert data[0, 0].real == 1.
    assert data[0, 0].imag == -1.
    assert data[0, 1].real == 2.
    assert data[0, 1].imag == -2.
    assert data[1, 0].real == 1
    assert data[1, 0].imag == -1
    check_simple_roundtrip(dic, data)


def test_2d_freq():
    """ read/write of 2D NMRPipe frequency domain file. """
    dic, data = ng.pipe.read(NMRPIPE_2D_FREQ)
    assert data.shape == (2, 8)
    assert data.dtype == 'float32'
    assert data[0, 0] == 1.
    assert data[0, 1] == 2.
    assert data[1, 0] == 1.
    check_simple_roundtrip(dic, data)
    check_ppm_limits(dic, data, 0, [179.00, 99.00])
    check_ppm_limits(dic, data, 1, [54.70, -32.80])


def test_2d_time_tp():
    """ read/write of 2D NMRPipe time domain file transposed. """
    dic, data = ng.pipe.read(NMRPIPE_2D_TIME_TP)
    assert data.shape == (16, 2)
    assert data.dtype == 'complex64'
    assert data[0, 0].real == 1.
    assert data[0, 0].imag == 1.
    assert data[0, 1].real == 1.
    assert data[0, 1].imag == 1.
    assert data[1, 0].real == -1.
    assert data[1, 0].imag == -1.
    assert data[2, 0].real == 2.
    assert data[2, 0].imag == 2.
    check_simple_roundtrip(dic, data)


def test_2d_freq_tp():
    """ read/write of 2D NMRPipe frequency domain file tranposed. """
    dic, data = ng.pipe.read(NMRPIPE_2D_FREQ_TP)
    assert data.shape == (8, 2)
    assert data.dtype == 'float32'
    assert data[0, 0] == 1.
    assert data[0, 1] == 1.
    assert data[1, 0] == 2.
    check_simple_roundtrip(dic, data)
    check_ppm_limits(dic, data, 0, [54.70, -32.80])
    check_ppm_limits(dic, data, 1, [179.00, 99.00])


def test_2d_time_lowmem():
    """ lowmem read/write of 2D NMRPipe time domain file. """
    dic, data = ng.pipe.read_lowmem(NMRPIPE_2D_TIME)
    assert data.shape == (4, 8)
    assert data.dtype == 'complex64'
    assert data[0, 0].real == 1.
    assert data[0, 0].imag == -1.
    assert data[0, 1].real == 2.
    assert data[0, 1].imag == -2.
    assert data[1, 0].real == 1
    assert data[1, 0].imag == -1
    check_simple_roundtrip(dic, data, lowmem=True)


def test_2d_freq_lowmem():
    """ lowmem read/write of 2D NMRPipe frequency domain file. """
    dic, data = ng.pipe.read_lowmem(NMRPIPE_2D_FREQ)
    assert data.shape == (2, 8)
    assert data.dtype == 'float32'
    assert data[0, 0] == 1.
    assert data[0, 1] == 2.
    assert data[1, 0] == 1.
    check_simple_roundtrip(dic, data, lowmem=True)
    check_ppm_limits(dic, data, 0, [179.00, 99.00])
    check_ppm_limits(dic, data, 1, [54.70, -32.80])


def test_3d_time():
    """ read/write of 3D NMRPipe time domain file. """
    dic, data = ng.pipe.read(NMRPIPE_3D_TIME)
    assert data.shape == (4, 6, 8)
    assert data.dtype == 'complex64'
    assert data[0, 0, 0].real == 1.
    assert data[0, 0, 0].imag == -1.
    assert data[0, 0, 1].real == 2.
    assert data[0, 0, 1].imag == -2.
    assert data[0, 1, 0].real == 1
    assert data[0, 1, 0].imag == -1
    assert data[1, 0, 0].real == 1
    assert data[1, 0, 0].imag == -1
    check_simple_roundtrip(dic, data, 1)


def test_3d_time_slice():
    """ read/write of 3D NMRPipe time domain first slice. """
    dic, data = ng.pipe.read(NMRPIPE_3D_TIME % (1))
    assert data.shape == (6, 8)
    assert data.dtype == 'complex64'
    assert data[0, 0].real == 1.
    assert data[0, 0].imag == -1.
    assert data[0, 1].real == 2.
    assert data[0, 1].imag == -2.
    assert data[1, 0].real == 1
    assert data[1, 0].imag == -1
    check_simple_roundtrip(dic, data)


def test_3d_freq():
    """ read/write of 3D NMRPipe frequency domain file. """
    dic, data = ng.pipe.read(NMRPIPE_3D_FREQ)
    assert data.shape == (2, 3, 8)
    assert data.dtype == 'float32'
    assert data[0, 0, 0] == 1.
    assert data[0, 0, 1] == 1.
    assert data[0, 1, 0] == 1.
    assert data[1, 0, 0] == 2.
    check_simple_roundtrip(dic, data, 1)
    check_ppm_limits(dic, data, 0, [220.00, 120.00])
    check_ppm_limits(dic, data, 1, [152.33, 45.67])
    check_ppm_limits(dic, data, 2, [54.70, -32.80])


def test_3d_freq_slice():
    """ read/write of 3D NMRPipe frequency domain first slice. """
    dic, data = ng.pipe.read(NMRPIPE_3D_FREQ % (1))
    assert data.shape == (3, 8)
    assert data.dtype == 'float32'
    assert data[0, 0] == 1.
    assert data[0, 1] == 1.
    assert data[1, 0] == 1.
    check_simple_roundtrip(dic, data)
    check_ppm_limits(dic, data, 0, [152.33, 45.67])
    check_ppm_limits(dic, data, 1, [54.70, -32.80])


def test_3d_time_lowmem():
    """ lowmem read/write of 3D NMRPipe time domain file. """
    dic, data = ng.pipe.read_lowmem(NMRPIPE_3D_TIME)
    assert data.shape == (4, 6, 8)
    assert data.dtype == 'complex64'
    assert data[0, 0, 0].real == 1.
    assert data[0, 0, 0].imag == -1.
    assert data[0, 0, 1].real == 2.
    assert data[0, 0, 1].imag == -2.
    assert data[0, 1, 0].real == 1
    assert data[0, 1, 0].imag == -1
    assert data[1, 0, 0].real == 1
    assert data[1, 0, 0].imag == -1
    check_simple_roundtrip(dic, data, 1, lowmem=True)


def test_3d_freq_lowmem():
    """ lowmem read/write of 3D NMRPipe frequency domain file. """
    dic, data = ng.pipe.read_lowmem(NMRPIPE_3D_FREQ)
    assert data.shape == (2, 3, 8)
    assert data.dtype == 'float32'
    assert data[0, 0, 0] == 1.
    assert data[0, 0, 1] == 1.
    assert data[0, 1, 0] == 1.
    assert data[1, 0, 0] == 2.
    check_simple_roundtrip(dic, data, 1, lowmem=True)
    check_ppm_limits(dic, data, 0, [220.00, 120.00])
    check_ppm_limits(dic, data, 1, [152.33, 45.67])
    check_ppm_limits(dic, data, 2, [54.70, -32.80])


def test_3d_time_stream():
    """ read/write of 3D NMRPipe time domain stream (single file). """
    dic, data = ng.pipe.read(NMRPIPE_3D_TIME_STREAM)
    assert data.shape == (4, 6, 8)
    assert data.dtype == 'complex64'
    assert data[0, 0, 0].real == 1.
    assert data[0, 0, 0].imag == -1.
    assert data[0, 0, 1].real == 2.
    assert data[0, 0, 1].imag == -2.
    assert data[0, 1, 0].real == 1
    assert data[0, 1, 0].imag == -1
    assert data[1, 0, 0].real == 1
    assert data[1, 0, 0].imag == -1
    check_simple_roundtrip(dic, data)


def test_3d_time_stream_lowmem():
    """ lowmem read/write of 3D NMRPipe time domain stream. """
    dic, data = ng.pipe.read_lowmem(NMRPIPE_3D_TIME_STREAM)
    assert data.shape == (4, 6, 8)
    assert data.dtype == 'complex64'
    assert data[0, 0, 0].real == 1.
    assert data[0, 0, 0].imag == -1.
    assert data[0, 0, 1].real == 2.
    assert data[0, 0, 1].imag == -2.
    assert data[0, 1, 0].real == 1
    assert data[0, 1, 0].imag == -1
    assert data[1, 0, 0].real == 1
    assert data[1, 0, 0].imag == -1
    check_simple_roundtrip(dic, data, lowmem=True)


def test_3d_freq_stream():
    """ read/write of 3D NMRPipe frequency domain stream. """
    dic, data = ng.pipe.read(NMRPIPE_3D_FREQ_STREAM)
    assert data.shape == (2, 3, 8)
    assert data.dtype == 'float32'
    assert data[0, 0, 0] == 1.
    assert data[0, 0, 1] == 1.
    assert data[0, 1, 0] == 1.
    assert data[1, 0, 0] == 2.
    check_simple_roundtrip(dic, data)
    check_ppm_limits(dic, data, 0, [220.00, 120.00])
    check_ppm_limits(dic, data, 1, [152.33, 45.67])
    check_ppm_limits(dic, data, 2, [54.70, -32.80])


def test_3d_freq_stream_lowmem():
    """ lowmem read/write of 3D NMRPipe frequency domain stream. """
    dic, data = ng.pipe.read_lowmem(NMRPIPE_3D_FREQ_STREAM)
    assert data.shape == (2, 3, 8)
    assert data.dtype == 'float32'
    assert data[0, 0, 0] == 1.
    assert data[0, 0, 1] == 1.
    assert data[0, 1, 0] == 1.
    assert data[1, 0, 0] == 2.
    check_simple_roundtrip(dic, data, lowmem=True)
    check_ppm_limits(dic, data, 0, [220.00, 120.00])
    check_ppm_limits(dic, data, 1, [152.33, 45.67])
    check_ppm_limits(dic, data, 2, [54.70, -32.80])


# All 4D tests are lowmem reads and writes

def test_4d_time_double_index():
    """ read/write of 4D NMRPipe time domain file (two index). """
    dic, data = ng.pipe.read_lowmem(NMRPIPE_4D_TIME_2)
    assert data.shape == (4, 6, 8, 5)
    assert data.dtype == 'complex64'
    assert data[0, 0, 0, 0].real == 1.
    assert data[0, 0, 0, 0].imag == -1.
    assert data[0, 0, 0, 1].real == 2.
    assert data[0, 0, 0, 1].imag == -2.
    assert data[0, 0, 1, 0].real == 1
    assert data[0, 0, 1, 0].imag == -1
    assert data[0, 1, 0, 0].real == 1
    assert data[0, 1, 0, 0].imag == -1
    assert data[1, 0, 0, 0].real == 1
    assert data[1, 0, 0, 0].imag == -1
    check_simple_roundtrip(dic, data, 2, lowmem=True)


def test_4d_time_single_index():
    """ read/write of 4D NMRPipe time domain file (single index). """
    dic, data = ng.pipe.read_lowmem(NMRPIPE_4D_TIME_1)
    assert data.shape == (4, 6, 8, 5)
    assert data.dtype == 'complex64'
    assert data[0, 0, 0, 0].real == 1.
    assert data[0, 0, 0, 0].imag == -1.
    assert data[0, 0, 0, 1].real == 2.
    assert data[0, 0, 0, 1].imag == -2.
    assert data[0, 0, 1, 0].real == 1
    assert data[0, 0, 1, 0].imag == -1
    assert data[0, 1, 0, 0].real == 1
    assert data[0, 1, 0, 0].imag == -1
    assert data[1, 0, 0, 0].real == 1
    assert data[1, 0, 0, 0].imag == -1
    check_simple_roundtrip(dic, data, 1, lowmem=True)


def test_4d_time_stream():
    """ read/write of 4D NMRPipe stream file. """
    dic, data = ng.pipe.read_lowmem(NMRPIPE_4D_TIME_STREAM)
    assert data.shape == (4, 6, 8, 5)
    assert data.dtype == 'complex64'
    assert data[0, 0, 0, 0].real == 1.
    assert data[0, 0, 0, 0].imag == -1.
    assert data[0, 0, 0, 1].real == 2.
    assert data[0, 0, 0, 1].imag == -2.
    assert data[0, 0, 1, 0].real == 1
    assert data[0, 0, 1, 0].imag == -1
    assert data[0, 1, 0, 0].real == 1
    assert data[0, 1, 0, 0].imag == -1
    assert data[1, 0, 0, 0].real == 1
    assert data[1, 0, 0, 0].imag == -1
    check_simple_roundtrip(dic, data, lowmem=True)


def test_4d_time_double_index_slice():
    """ read/write of 4D NMRPipe time domain (two index) first slice. """
    dic, data = ng.pipe.read(NMRPIPE_4D_TIME_2 % (1, 1))
    assert data.shape == (8, 5)
    assert data.dtype == 'complex64'
    assert data[0, 0].real == 1.
    assert data[0, 0].imag == -1.
    assert data[0, 1].real == 2.
    assert data[0, 1].imag == -2.
    assert data[1, 0].real == 1
    assert data[1, 0].imag == -1
    check_simple_roundtrip(dic, data)


def test_4d_time_single_index_slice():
    """ read/write of 4D NMRPipe time domain (single index) first slice. """
    dic, data = ng.pipe.read(NMRPIPE_4D_TIME_1 % (1))
    assert data.shape == (8, 5)
    assert data.dtype == 'complex64'
    assert data[0, 0].real == 1.
    assert data[0, 0].imag == -1.
    assert data[0, 1].real == 2.
    assert data[0, 1].imag == -2.
    assert data[1, 0].real == 1
    assert data[1, 0].imag == -1
    check_simple_roundtrip(dic, data)


def test_4d_freq_single_index():
    """ read/write of 4D NMRPipe frequency domain file (single index). """
    dic, data = ng.pipe.read_lowmem(NMRPIPE_4D_FREQ_1)
    assert data.shape == (2, 3, 4, 5)
    assert data.dtype == 'float32'
    assert data[0, 0, 0, 0] == 1.
    assert data[0, 0, 0, 1] == 1.
    assert data[0, 0, 1, 0] == 1.
    assert data[0, 1, 0, 0] == 1.
    assert data[1, 0, 0, 0] == 2.
    check_ppm_limits(dic, data, 0, [180.00, 80.00])
    check_ppm_limits(dic, data, 1, [186.67, 53.33])
    check_ppm_limits(dic, data, 2, [179.00, 59.00])
    check_ppm_limits(dic, data, 3, [44.70, -35.30])
    check_simple_roundtrip(dic, data, 1, lowmem=True)


def test_4d_freq_double_index():
    """ read/write of 4D NMRPipe frequency domain file (double index). """
    dic, data = ng.pipe.read_lowmem(NMRPIPE_4D_FREQ_2)
    assert data.shape == (2, 3, 4, 5)
    assert data.dtype == 'float32'
    assert data[0, 0, 0, 0] == 1.
    assert data[0, 0, 0, 1] == 1.
    assert data[0, 0, 1, 0] == 1.
    assert data[0, 1, 0, 0] == 1.
    assert data[1, 0, 0, 0] == 2.
    check_ppm_limits(dic, data, 0, [180.00, 80.00])
    check_ppm_limits(dic, data, 1, [186.67, 53.33])
    check_ppm_limits(dic, data, 2, [179.00, 59.00])
    check_ppm_limits(dic, data, 3, [44.70, -35.30])
    check_simple_roundtrip(dic, data, 2, lowmem=True)


def test_4d_freq_stream():
    """ read/write of 4D NMRPipe frequency domain stream file. """
    dic, data = ng.pipe.read_lowmem(NMRPIPE_4D_FREQ_STREAM)
    assert data.shape == (2, 3, 4, 5)
    assert data.dtype == 'float32'
    assert data[0, 0, 0, 0] == 1.
    assert data[0, 0, 0, 1] == 1.
    assert data[0, 0, 1, 0] == 1.
    assert data[0, 1, 0, 0] == 1.
    assert data[1, 0, 0, 0] == 2.
    check_ppm_limits(dic, data, 0, [180.00, 80.00])
    check_ppm_limits(dic, data, 1, [186.67, 53.33])
    check_ppm_limits(dic, data, 2, [179.00, 59.00])
    check_ppm_limits(dic, data, 3, [44.70, -35.30])
    check_simple_roundtrip(dic, data, lowmem=True)


def test_4d_freq_double_index_slice():
    """ read/write of 4D NMRPipe freq domain (two index) first slice. """
    dic, data = ng.pipe.read(NMRPIPE_4D_FREQ_2 % (1, 1))
    assert data.shape == (4, 5)
    assert data.dtype == 'float32'
    assert data[0, 0] == 1.
    assert data[0, 1] == 1.
    assert data[1, 0] == 1.
    check_ppm_limits(dic, data, 0, [179.00, 59.00])
    check_ppm_limits(dic, data, 1, [44.70, -35.30])
    check_simple_roundtrip(dic, data)


def test_4d_freq_single_index_slice():
    """ read/write of 4D NMRPipe freq domain (single index) first slice. """
    dic, data = ng.pipe.read(NMRPIPE_4D_FREQ_1 % (1))
    assert data.shape == (4, 5)
    assert data.dtype == 'float32'
    assert data[0, 0] == 1.
    assert data[0, 1] == 1.
    assert data[1, 0] == 1.
    check_ppm_limits(dic, data, 0, [179.00, 59.00])
    check_ppm_limits(dic, data, 1, [44.70, -35.30])
    check_simple_roundtrip(dic, data)


def test_make_uc():
    """ ng.pipe.make_uc """
    dic, data = ng.pipe.read(NMRPIPE_2D_FREQ)

    # H1 dimension
    uc = ng.pipe.make_uc(dic, data)
    assert uc._size == 8
    assert uc._cplx is False
    assert uc._sw == 50000.0
    assert uc._obs == 500.0
    assert round(uc._car) == 2350.

    # H1 dimension
    uc = ng.pipe.make_uc(dic, data, 1)
    assert uc._size == 8
    assert uc._cplx is False
    assert uc._sw == 50000.0
    assert uc._obs == 500.0
    assert round(uc._car) == 2350.

    # C13 dimension
    uc = ng.pipe.make_uc(dic, data, 0)
    assert uc._size == 2
    assert uc._cplx is False
    assert uc._sw == 20000.0
    assert uc._obs == 125.0
    assert round(uc._car) == 12375.


def test_guess_udic():
    """ ng.pipe.guess_udic """
    dic, data = ng.pipe.read(NMRPIPE_2D_FREQ)
    udic = ng.pipe.guess_udic(dic, data)
    assert udic[0]['car'] == 12375.0
    assert udic[0]['complex'] is False
    assert udic[0]['encoding'] == 'states'
    assert udic[0]['freq'] is True
    assert udic[0]['label'] == 'C13'
    assert udic[0]['obs'] == 125.0
    assert udic[0]['size'] == 2
    assert udic[0]['sw'] == 20000.0
    assert udic[0]['time'] is False
    assert round(udic[1]['car']) == 2350
    assert udic[1]['complex'] is False
    assert udic[1]['encoding'] == 'states'
    assert udic[1]['freq'] is True
    assert udic[1]['label'] == 'H1'
    assert udic[1]['obs'] == 500.0
    assert udic[1]['size'] == 8
    assert udic[1]['sw'] == 50000.0
    assert udic[1]['time'] is False
    assert udic['ndim'] == 2

    dic, data = ng.pipe.read(NMRPIPE_2D_TIME)
    udic = ng.pipe.guess_udic(dic, data)
    assert udic[0]['car'] == 12375.0
    assert udic[0]['complex'] is True
    assert udic[0]['encoding'] == 'states'
    assert udic[0]['freq'] is False
    assert udic[0]['label'] == 'C13'
    assert udic[0]['obs'] == 125.0
    assert udic[0]['size'] == 4
    assert udic[0]['sw'] == 20000.0
    assert udic[0]['time'] is True
    assert round(udic[1]['car']) == 2350
    assert udic[1]['complex'] is True
    assert udic[1]['encoding'] == 'states'
    assert udic[1]['freq'] is False
    assert udic[1]['label'] == 'H1'
    assert udic[1]['obs'] == 500.0
    assert udic[1]['size'] == 8
    assert udic[1]['sw'] == 50000.0
    assert udic[1]['time'] is True
    assert udic['ndim'] == 2


def test_read_table():
    comments, tbl_format, tbl = ng.pipe.read_table(NMRPIPE_TABLE)

    assert len(comments) == 13
    assert len(tbl_format) == 25
    assert len(tbl) == 5
    assert len(tbl[0]) == 25

    assert tbl_format[0] == '%5d'
    assert tbl_format[1] == '%9.3f'

    assert tbl['INDEX'][0] == 1
    assert tbl['INDEX'][-1] == 5

    assert tbl['X1'][0] == 1328
    assert tbl['X1'][-1] == 1161


def test_write_table():
    ref_comments, ref_tbl_format, ref_tbl = ng.pipe.read_table(NMRPIPE_TABLE)
    try:
        tbl_fname = tempfile.mktemp(dir='.')
        ng.pipe.write_table(tbl_fname, ref_comments, ref_tbl_format, ref_tbl)
        comments, tbl_format, tbl = ng.pipe.read_table(tbl_fname)
        assert all(ref_tbl == tbl)
        assert ref_comments == comments
        assert ref_tbl_format == tbl_format
    finally:
        os.remove(tbl_fname)



"""
Functions to perform segmentation of NMR spectrum.
"""

import numpy as np
import numpy.ma as ma
import scipy.ndimage as ndimage

from .analysisbase import neighbors

# Connected segmenting method:
# The connected segmentation method finds all nodes which are above a given
# threshold and connected to the initial point.  For finding all segments
# the scipy.ndimage.label function is used for speed.


def label_connected(data, thres, structure):
    """
    Label connected features in data.  Returns labeled_array, num_features
    """
    return ndimage.label(data > thres, structure)


def find_all_connected(data, thres, find_segs=False, diag=False):
    """
    Find all connected segments.

    Parameters
    ----------
    data : ndarray
        Data to perform segmentation on.
    thres : float
        Threshold, below this nodes are considered noise.
    find_segs : bool, optional
        True to return a list of slices for the segments.
    diag : bool
        True to include diagonal neighbors in connection.

    Returns
    -------
    locations : list
        List of indicies of local maximum in each segment.
    seg_slices : list, optional
        List of slices which extract a given segment from the data. Only
        returned when fig_segs is True.

    """
    # build structure array for defining feature connections
    ndim = data.ndim
    if diag:
        structure = ndimage.generate_binary_structure(ndim, ndim)
    else:
        structure = ndimage.generate_binary_structure(ndim, 1)

    # determine labeled array of segments
    labels, num_features = label_connected(data, thres, structure)

    # determine locations of segment maxima
    locations = ndimage.maximum_position(data, labels, range(1, num_features +
                                                             1))
    # find segment slices if requested and return
    if find_segs is True:
        seg_slices = ndimage.find_objects(labels)
        return locations, seg_slices
    else:
        return locations

# nconnected method:
# The nconnected method is identical to the connected method except nodes must
# be below the threshold and local minimum are reported.  This is useful for
# finding negative peaks by setting thres to the negative of the noise level.


def label_nconnected(data, thres, structure):
    """
    Label nconnected features in data.  Returns labeled_array, num_features
    """
    return ndimage.label(data < thres, structure)


def find_all_nconnected(data, thres, find_segs=False, diag=False):
    """
    Find all negatively connected segments in data.

    Parameters
    ----------
    data : ndarray
        Data to perform segmentation on.
    thres : float
        Threshold, below this nodes are considered noise.
    find_segs : bool, optional
        True to return a list of slices for the segments.
    diag : bool
        True to include diagonal neighbors in connection.

    Returns
    -------
    locations : list
        List of indicies of local maximum in each segment.
    seg_slices : list, optional
        List of slices which extract a given segment from the data. Only
        returned when fig_segs is True.

    """
    # build structure array for defining feature connections
    ndim = data.ndim
    if diag:
        structure = ndimage.generate_binary_structure(ndim, ndim)
    else:
        structure = ndimage.generate_binary_structure(ndim, 1)

    # determine labeled array of segments
    labels, num_features = label_nconnected(data, thres, structure)

    # determine locations of segment maxima
    locations = ndimage.minimum_position(data, labels, range(1,
                                         num_features + 1))
    # find segment slices if requested and return
    if find_segs is True:
        seg_slices = ndimage.find_objects(labels)
        return locations, seg_slices
    else:
        return locations

# downward segmentation method:
# The downward segmenting method uses the flood fill algorithm to find
# all points connected to an initial node which are above a given threshold
# and to which a path exists in which each step of the path moves lower in
# intensity.  This can be though of as all points accessible by a water drop
# following downward slopes from the initial node.

# Upward segmentation uses the same priciple except nodes must be below
# the threshold an upward path must exist.


def mark_dseg(mdata, map, pt, mark, structure):
    """
    Mark downward-connected region on segment map starting at node pt.

    Modifies mdata mask and map.

    Parameters
    ----------
    mdata : masked ndarray
        Masked data array.
    map :
        Array mapping out segments.
    pt : tuple of ints
        Index of starting node
    mark : int
        Integer to mark map with.

    """
    if mdata.mask[pt] is True:
        return
    else:
        map[pt] = mark
        mdata[pt] = ma.masked

    Q = [pt]
    while Q:
        pt = Q.pop(0)
        v = mdata.data[pt]
        # Check all neightbors
        for new_pt in neighbors(pt, mdata.shape, structure):
            if mdata.mask[new_pt] == False and mdata[new_pt] < v:
                Q.append(new_pt)
                map[new_pt] = mark
                mdata[new_pt] = ma.masked
    return


def label_downward_seg(data, labels, seg_slice, seg_index, max_index,
                       structure):
    """ Label a segment which is downward connected """
    slabels = labels[seg_slice]
    msdata = np.ma.masked_array(data[seg_slice], mask=(slabels != seg_index))

    # mark the downward connected segment with the highest peak in the
    # selected segment with the segment index.
    argmax = np.unravel_index(msdata.argmax(), msdata.shape)
    mark_dseg(msdata, slabels, argmax, seg_index, structure)

    # mark any
    while msdata.mask.all() == False:
        argmax = np.unravel_index(msdata.argmax(), msdata.shape)
        mark_dseg(msdata, slabels, argmax, max_index, structure)
        max_index = max_index + 1
    return max_index


def label_downward(data, thres, structure):
    """
    Label connected features in data. Returns labeled_array, num_features
    """
    # find connected segments
    labels, num_features = ndimage.label(data > thres, structure)
    seg_slices = ndimage.find_objects(labels)
    max_index = int(num_features + 1)

    # loop over the segments and perform a downward segment on each
    for i, s in enumerate(seg_slices):
        max_index = label_downward_seg(data, labels, s, i + 1, max_index,
                                       structure)
    return labels, max_index - 1


def find_all_downward(data, thres, find_segs=False, diag=False):
    """
    Find all downward connected segments in data

    Parameters
    ----------
    data : ndarray
        Data to perform segmentation on.
    thres : float
        Threshold, below this nodes are considered noise.
    find_segs : bool, optional
        True to return a list of slices for the segments.
    diag : bool, optional
        True to include diagonal neighbors in connection.

    Returns
    -------
    locations : list
        List of indicies of local maximum in each segment.
    seg_slices : list, optional
        List of slices which extract a given segment from the data. Only
        returned when fig_segs is True.

    """
    # build structure array for defining feature connections
    ndim = data.ndim
    if diag:
        structure = ndimage.generate_binary_structure(ndim, ndim)
    else:
        structure = ndimage.generate_binary_structure(ndim, 1)

    # determine labeled array of segments
    labels, num_features = label_downward(data, thres, structure)

    # determine locations of segment maxima
    locations = ndimage.maximum_position(data, labels, range(1,
                                         num_features + 1))

    # find segment slices if requested and return
    if find_segs is True:
        seg_slices = ndimage.find_objects(labels)
        return locations, seg_slices
    else:
        return locations


def mark_useg(mdata, map, pt, mark, structure):
    """
    Mark upward-connected region on segment map starting at node pt

    Modifies mdata mask and map.

    Parameters
    ----------
    mdata : masked ndarray
        Masked data array.
    map :
        Array mapping out segments.
    pt : tuple of ints
        Index of starting node
    mark : int
        Integer to mark map with.

    """
    if mdata.mask[pt] is True:
        return
    else:
        map[pt] = mark
        mdata[pt] = ma.masked

    Q = [pt]
    while Q:
        pt = Q.pop(0)
        v = mdata.data[pt]
        # Check all neightbors
        for new_pt in neighbors(pt, mdata.shape, structure):
            if mdata.mask[new_pt] == False and mdata[new_pt] > v:
                Q.append(new_pt)
                map[new_pt] = mark
                mdata[new_pt] = ma.masked
    return


def label_upward_seg(data, labels, seg_slice, seg_index, max_index,
                     structure):
    """ Label a segment which is upward connected """
    slabels = labels[seg_slice]
    msdata = np.ma.masked_array(data[seg_slice],
                                mask=(slabels != seg_index))
    # mark the upward connected segment with the highest peak in the
    # selected segment with the segment index.
    argmin = np.unravel_index(msdata.argmin(), msdata.shape)
    mark_useg(msdata, slabels, argmin, seg_index, structure)

    # mark any
    while msdata.mask.all() == False:
        argmin = np.unravel_index(msdata.argmin(), msdata.shape)
        mark_useg(msdata, slabels, argmin, max_index, structure)
        max_index = max_index + 1

    return max_index


def label_upward(data, thres, structure):
    """
    Label upward connected features in data. Returns labeled_array,
    num_features
    """
    # find connected segments
    labels, num_features = ndimage.label(data < thres, structure)
    seg_slices = ndimage.find_objects(labels)
    max_index = int(num_features + 1)

    # loop over the segments and perform a downward segment on each
    for i, s in enumerate(seg_slices):
        max_index = label_upward_seg(data, labels, s, i + 1, max_index,
                                     structure)
    return labels, max_index - 1


def find_all_upward(data, thres, find_segs=False, diag=False):
    """
    Find all upward connected segments in data

    Parameters
    ----------
    data : ndarray
        Data to perform segmentation on.
    thres : float
        Threshold, below this nodes are considered noise.
    find_segs : bool, optional
        True to return a list of slices for the segments.
    diag : bool, optional
        True to include diagonal neighbors in connection.

    Returns
    -------
    locations : list
        List of indicies of local maximum in each segment.
    seg_slices : list, optional
        List of slices which extract a given segment from the data. Only
        returned when fig_segs is True.

    """
    # build structure array for defining feature connections
    ndim = data.ndim
    if diag:
        structure = ndimage.generate_binary_structure(ndim, ndim)
    else:
        structure = ndimage.generate_binary_structure(ndim, 1)

    # determine labeled array of segments
    labels, num_features = label_upward(data, thres, structure)

    # determine locations of segment maxima
    locations = ndimage.minimum_position(data, labels,
                                         range(1, num_features + 1))
    # find segment slices if requested and return
    if find_segs is True:
        seg_slices = ndimage.find_objects(labels)
        return locations, seg_slices
    else:
        return locations


##########################
# Single point functions #
##########################


def find_downward(data, pt, thres, diag=False):
    """
    Find points downward-connected to a point in data.

    Parameters
    ----------
    data : ndarray
        Array of data.
    pt : tuple of ints
        Starting point of peak.
    thres : float
        Threshold, below this nodes are considered noise.
    diag : bool, optional
        True to include diagonal neighbors in connection.

    Returns
    -------
    nodes : list
        Indicies of downward-connected nodes.

    """
    # build structure array for defining feature connections
    ndim = data.ndim
    if diag:
        structure = ndimage.generate_binary_structure(ndim, ndim)
    else:
        structure = ndimage.generate_binary_structure(ndim, 1)

    if isinstance(pt, int):
        pt = (pt, )
    pt = tuple(pt)
    shape = data.shape

    if data[pt] < thres:    # check that the initial point is above threshold.
        return []
    Q = [pt]    # queue
    segment = [pt]

    while Q:    # loop until Q is empty
        pt = Q.pop(0)   # remove first element of queue
        v = data[pt]    # value at current node
        for new_pt in neighbors(pt, shape, structure):  # check all neightbors
            if thres < data[new_pt] < v and new_pt not in segment:
                Q.append(new_pt)
                segment.append(new_pt)
    return segment


def find_connected(data, pt, thres, diag=False):
    """
    Find points connected to a point in data.

    Parameters
    ----------
    data : ndarray
        Array of data.
    pt : tuple of ints
        Starting point of peak.
    thres : float
        Threshold, below this nodes are considered noise.
    diag : bool, optional
        True to include diagonal neighbors in connection.

    Returns
    -------
    nodes : list
        Indicies of connected nodes.

    """
    # build structure array for defining feature connections
    ndim = data.ndim
    if diag:
        structure = ndimage.generate_binary_structure(ndim, ndim)
    else:
        structure = ndimage.generate_binary_structure(ndim, 1)

    if isinstance(pt, int):
        pt = (pt, )
    pt = tuple(pt)
    shape = data.shape

    if data[pt] < thres:    # check that the initial point is above threshold.
        return []
    Q = [pt]    # queue
    segment = [pt]

    while Q:    # loop until Q is empty
        pt = Q.pop(0)   # remove first element of queue
        for new_pt in neighbors(pt, shape, structure):  # check all neightbors
            if data[new_pt] > thres and new_pt not in segment:
                Q.append(new_pt)
                segment.append(new_pt)
    return segment


def find_nconnected(data, pt, thres, diag=False):
    """
    Find points connected to pt in data below threshold.

    Parameters
    ----------
    data : ndarray
        Array of data.
    pt : tuple of ints
        Starting point of peak.
    thres : float
        Threshold, above this nodes are considered noise.
    diag : bool, optional
        True to include diagonal neighbors in connection.

    Returns
    -------
    nodes : list
        Indicies of connected nodes.

    """
    # build structure array for defining feature connections
    ndim = data.ndim
    if diag:
        structure = ndimage.generate_binary_structure(ndim, ndim)
    else:
        structure = ndimage.generate_binary_structure(ndim, 1)

    if isinstance(pt, int):
        pt = (pt, )
    pt = tuple(pt)
    shape = data.shape

    if data[pt] > thres:    # check that the initial point is above threshold.
        return []
    Q = [pt]    # queue
    segment = [pt]

    while Q:    # loop until Q is empty
        pt = Q.pop(0)   # remove first element of queue
        for new_pt in neighbors(pt, shape, structure):  # check all neightbors
            if data[new_pt] < thres and new_pt not in segment:
                Q.append(new_pt)
                segment.append(new_pt)
    return segment


def find_upward(data, pt, thres, diag=False):
    """
    Find points upward-connected to pt in data.

    Parameters
    ----------
    data : ndarray
        Array of data.
    pt : tuple of ints
        Starting point of peak.
    thres : float
        Threshold, below this nodes are considered noise.
    diag : bool, optional
        True to include diagonal neighbors in connection.

    Returns
    -------
    nodes : list
        Indicies of upward-connected nodes.

    """
    # build structure array for defining feature connections
    ndim = data.ndim
    if diag:
        structure = ndimage.generate_binary_structure(ndim, ndim)
    else:
        structure = ndimage.generate_binary_structure(ndim, 1)

    if isinstance(pt, int):
        pt = (pt, )
    pt = tuple(pt)
    shape = data.shape

    if data[pt] > thres:    # check that the initial point is below threshold.
        return []
    Q = [pt]    # queue
    segment = [pt]

    while Q:    # loop until Q is empty
        pt = Q.pop(0)   # remove first element of queue
        v = data[pt]    # value at current node
        for new_pt in neighbors(pt, shape, structure):  # check all neightbors
            if thres > data[new_pt] > v and new_pt not in segment:
                Q.append(new_pt)
                segment.append(new_pt)
    return segment
